from datetime import datetime
from typing import Annotated, Dict, List, Optional, Tuple

from pydantic import BaseModel, Field
from pytest import raises

from melvin.tecton_gen_ai.utils._serialization import (
    to_openai_properties,
    openai_function_to_pydantic,
    openai_properties_to_pydantic,
    openai_type_to_python_type,
    to_openai_function,
)


class Input1(BaseModel):
    """
    hello
    world"""

    a: str = Field(description="a description")
    b: int = Field(description="b description")
    c: float = Field(description="c description", default=0.1)
    d: bool = Field(description="d description")
    e: List[str] = Field(description="e description")
    f: Dict[str, int] = Field(description="f description")
    g: datetime = Field(description="g description")
    h: List[Tuple[str, str]] = Field(description="h description", default=[])
    i: Optional[List[Tuple[str, str]]] = Field(description="i description")


class Input2(BaseModel):
    pass


class Input3(BaseModel):
    """hello"""

    a: str


class Output(BaseModel):
    a: int = Field(description="a description")


def test_to_openai_function():
    res = to_openai_function(Input1, "test1")
    assert res["description"] == "\nhello\nworld"
    model = openai_function_to_pydantic(res, name="Input1")
    assert model.__doc__ == "\nhello\nworld"
    _assert_eq(Input1, model)

    def f(a: str) -> int:
        """hello"""
        pass

    res = openai_function_to_pydantic(to_openai_function(f, "test1"), name="Input3")
    _assert_eq(Input3, res)


def test_openai_type_to_python_type():
    assert openai_type_to_python_type({"type": "integer"}) is int
    assert openai_type_to_python_type({"type": "number"}) is float
    assert openai_type_to_python_type({"type": "boolean"}) is bool
    assert openai_type_to_python_type({"type": "string"}) is str
    assert (
        openai_type_to_python_type({"type": "array", "items": {"type": "string"}})
        == List[str]
    )
    assert (
        openai_type_to_python_type(
            {"type": "object", "additionalProperties": {"type": "integer"}}
        )
        == Dict[str, int]
    )
    assert (
        openai_type_to_python_type({"type": "string", "format": "date-time"})
        == datetime
    )
    with raises(NotImplementedError):
        openai_type_to_python_type({"type": "string", "format": "date"})


def test_extract_parameters_into_openai_properties():
    class Dummy:
        pass

    def f(
        a: str,
        b: Dummy,
        c: Annotated[int, Field(description="hello")],
        d: Annotated[str, "world"],
    ):
        pass

    res = to_openai_properties(f, ["a", "c", "d"])
    actual = openai_properties_to_pydantic(res, name="Expected")

    class Expected(BaseModel):
        a: str
        c: int = Field(description="hello")
        d: str = Field(description="world")

    _assert_eq(Expected, actual)

    class Expected2(BaseModel):
        a: str
        d: str = Field(description="world")

    res = to_openai_properties(Expected, ["a", "d"])
    actual = openai_properties_to_pydantic(res, name="Expected2")

    _assert_eq(Expected2, actual)

    res = to_openai_properties([("a", str), ("d", Annotated[str, "world"])], ["a", "d"])
    actual = openai_properties_to_pydantic(res, name="Expected2")

    _assert_eq(Expected2, actual)


def _assert_eq(a: BaseModel, b: BaseModel):
    expected = a.model_json_schema()
    del expected["title"]
    actual = b.model_json_schema()
    del actual["title"]
    assert expected == actual



import json

import jsonschema
from pydantic.v1.datetime_parse import parse_datetime
from tecton.types import Array, Bool, Field, Float64, Int64, String, Struct, Timestamp


def _get_type(schema_field, schema_dict):
    if "$ref" in schema_field:
        obj = _get_obj_from_ref(schema_field["$ref"], schema_dict)
        return _get_type(obj, schema_dict)

    if "type" not in schema_field:
        raise ValueError(f"Type not found: {schema_field}")

    dtype_str = schema_field["type"]
    if dtype_str == "string":
        if "format" in schema_field:
            if schema_field["format"] == "date-time":
                return Timestamp
            else:
                raise ValueError(f"Format {schema_field['format']} not supported")
        else:
            return String
    elif dtype_str == "number":
        return Float64
    elif dtype_str == "integer":
        return Int64
    elif dtype_str == "boolean":
        return Bool
    elif dtype_str == "array":
        item_type = _get_type(schema_field["items"], schema_dict)
        return Array(item_type)
    elif dtype_str == "object":
        struct_fields = _process_object(schema_field["properties"], schema_dict)
        return Struct(struct_fields)
    elif "$ref" in dtype_str:
        obj = _get_obj_from_ref(dtype_str["$ref"], schema_dict)
        return _get_type(obj, schema_dict)
    else:
        raise ValueError(f"Type {dtype_str} not supported")


def _get_obj_from_ref(ref, schema_dict):
    path = ref.split("/")[1:]
    obj = schema_dict
    for item in path:
        obj = obj[item]
    return obj


def _process_object(properties, schema_dict):
    fields = []
    for property_name, property_value in properties.items():
        if "$ref" in property_value:
            obj = _get_obj_from_ref(property_value["$ref"], schema_dict)

            if obj["type"] == "object":
                struct_fields = _process_object(obj["properties"], schema_dict)
                fields.append(Field(name=property_name, dtype=Struct(struct_fields)))
            else:
                fields.append(
                    Field(name=property_name, dtype=_get_type(obj, schema_dict))
                )
        else:
            fields.append(
                Field(name=property_name, dtype=_get_type(property_value, schema_dict))
            )
    return fields


def get_tecton_fields_from_json_schema(schema_dict):
    """Converts a JSON schema into a list of Tecton Field types"""
    struct = _get_type(schema_dict, schema_dict)
    return struct.fields


def _get_field_converter(schema, value):
    if schema.get("format") == "date-time" and isinstance(value, str):
        return parse_datetime
    return None


def _convert_string_format_recursive(obj, schema):
    if isinstance(obj, dict):
        for key, value in obj.items():
            if key in schema.get("properties", {}):
                prop_schema = schema["properties"][key]
                if converter := _get_field_converter(prop_schema, value):
                    obj[key] = converter(value)
                elif isinstance(value, (dict, list)):
                    _convert_string_format_recursive(value, prop_schema)
    elif isinstance(obj, list) and schema.get("items"):
        item_schema = schema["items"]
        for i, item in enumerate(obj):
            if converter := _get_field_converter(item_schema, item):
                obj[i] = converter(item)
            elif isinstance(item, (dict, list)):
                _convert_string_format_recursive(item, item_schema)


def load_to_rich_dict(json_str: str, schema):
    output = json.loads(json_str)
    jsonschema.validate(output, schema)
    _convert_string_format_recursive(output, schema)
    return output



import os
from typing import Any, List

from utils import (
    _err,
    _get_source_code,
    _log,
    _success,
    _code,
    get_cwd,
    is_in_feature_repo,
    run_command,
    validate_syntax,
)

# from tecton_utils import extract_tecton_objects


def get_source_code_lineno() -> List[List[Any]]:
    """
    Get the python code (with line numbers starting from 1) for all data source and feature definitions

    Returns:
        list[list[str]]: The python code for all feature definitions,
            each element is also a list. It contains the line number as the first element
            and the code as the second element.
    """
    assert is_in_feature_repo(), "You must be in a feature repo"

    path = os.path.join(get_cwd(), "features.py")
    if not os.path.exists(path):
        res = []
    else:
        with open(path, "r") as f:
            lines = f.read().splitlines()
            res = [[i + 1, line] for i, line in enumerate(lines)]
    if len(res) == 0:
        res = [[1, ""]]
    if res[-1][1] != "":
        res.append([len(res) + 1, ""])
    return res


def get_source_code() -> str:
    """
    Get the python source code of features.py

    Returns:

        str: The python code of features.py
    """
    assert is_in_feature_repo(), "You must be in a feature repo"

    path = os.path.join(get_cwd(), "features.py")
    _log(":eyeglasses: Getting the source code from " + path)
    return _get_source_code()


def modify_source_code(
    modifications: dict[int, str] = {},
    insertions: dict[int, list[str]] = {},
    deletions: dict[int, int] = {},
    explanation: str = "",
) -> str:
    """
    Modify the code of data source and feature definitions, before each call to this tool,
    get_source_code must be used to get the latest original code,
    so the delta change can be calculated based on that.

    Args:

        modifications: A dictionary with line number as key and the new code as value, provide empty dict if no modification
        insertions: A dictionary with insertion line number as key and a list of new lines as value, provide empty dict if no insertion
        deletions: A dictionary with start line number as key and count of lines to delete as value, provide empty dict if no deletion
        explanation: An explanation of the changes, in under 100 words

    Returns:

        str: a message indicating whether the change is successful

    Note:

        To modify the code, it is always helpful to search tecton documents.
    """
    return _modify_source_code(
        modifications=modifications,
        insertions=insertions,
        deletions=deletions,
        explanation=explanation,
        revert_on_error=False,
    )


def fix_syntax_issues(
    modifications: dict[int, str] = {},
    insertions: dict[int, list[str]] = {},
    deletions: dict[int, int] = {},
    explanation: str = "",
) -> str:
    """
    Modify the code to fix syntax issues.
    get_source_code must be used to get the latest original code,
    so the delta change can be calculated based on that.

    Args:

        modifications: A dictionary with line number as key and the new code as value, provide empty dict if no modification
        insertions: A dictionary with insertion line number as key and a list of new lines as value, provide empty dict if no insertion
        deletions: A dictionary with start line number as key and count of lines to delete as value, provide empty dict if no deletion
        explanation: An explanation of the changes, in under 100 words

    Returns:

        str: a message indicating whether the change is successful

    Note:

        If the modification caused new syntax error, the change will be reverted and you should retry.
    """
    path = os.path.join(get_cwd(), "features.py")
    try:
        validate_syntax(path)
        return _success("No syntax issues found")
    except SyntaxError:
        return _modify_source_code(
            modifications=modifications,
            insertions=insertions,
            deletions=deletions,
            explanation="Syntax fixer: " + explanation,
            revert_on_error=True,
        )
    except Exception as e:
        return _err(e)


def save_modified_code(
    python_code: str, explanation: str = "", previous_problem: str = ""
) -> str:
    """
    Save the new feature definition source code.
    The new code must be the whole content of the file.

    Args:

        python_code: The new python code for the features.py
        explanation: An explanation of the changes, in under 100 words
        previous_problem: A summary of the previous problem leading to this change. Set this to
            non-empty string only when the change is to fix an issure reported by a validation.
            The summary should be under 100 words. If non-empty, it should start with "Previous problem: "

    Returns:

        str: The success or error message
    """
    if previous_problem:
        _log(":exclamation: " + previous_problem)
    _log(":writing_hand: " + explanation)

    path = os.path.join(get_cwd(), "features.py")
    if not python_code.endswith("\n"):
        python_code += "\n"
    with open(path, "w") as f:
        f.write(python_code)
    try:
        validate_syntax(path)
        run_command(f"black {path}")
    except SyntaxError as se:
        return _err(se, prefix="Syntax error")
    finally:
        _code()
    return _success("Feature definitions updated without syntax issues")


def validate_with_tecton_plan() -> str:
    """
    Validate the current feature code using `tecton plan`

    Returns:

        str: The validation result
    """
    try:
        assert is_in_feature_repo(), "You must be in in a feature repo"
        _log(":question: Validating the code change")
        path = os.path.join(get_cwd(), "features.py")
        validate_syntax(path)
        # tecton_objects = extract_tecton_objects(path)
        code, out, err = run_command("tecton plan")
        if code == 0:
            return _success(
                "The code has been successfully validated via `tecton plan`. No issues found."
            )
        else:
            return _err(f"{err}\n\n{out}")
    except Exception as e:
        return _err(e)


def generate_flowchart_representation(
    nodes: list[str], edges: list[list[str]], types: list[str]
) -> str:
    """
    Generate a Tecton flowchart representation from the code or description in the context.
    It is for visualization purpose.

    Args:

        nodes: A list of node names
        edges: A list of edges, each edge must be a list of two node names in `nodes`
        types: A list of Tecton types, it must have the same length as `nodes`

    Returns:

        str: The flowchart representation.

    Note:

        - Each node must represent a Tecton object (e.g entity or feature view)
        - For a decorated function, node name is the functio name, and the type is the decorator name
        - The node names should be valid python identifiers
        - If there are errors, the return starts with "Error: "
        - The representation wiil ber surrounded by triple backticks, with the language set to "tecton_diagram"

    Example output:

    ```tecton_diagram
    {"rand_id": "gsdfg", "nodes": ["entity1", "feature_view1"], "edges": [["entity1", "feature_view1"]], "types": ["Entity", "BatchFeatureView"]}
    ```
    """
    import json
    import uuid

    rand_id = str(uuid.uuid4())[:5]

    repr = json.dumps(
        {"rand_id": rand_id, "nodes": nodes, "edges": edges, "types": types}
    )
    for edge in edges:
        if len(edge) != 2:
            return _err("Each edge must have two nodes: " + repr)
        if edge[0] not in nodes:
            return _err(f"Invalid source node {edge[0]} in an edge: " + repr)
        if edge[1] not in nodes:
            return _err(f"Invalid target node {edge[1]} in an edge: " + repr)
    return repr


def _modify_source_code(
    modifications: dict[int, str] = {},
    insertions: dict[int, list[str]] = {},
    deletions: dict[int, int] = {},
    explanation: str = "",
    revert_on_error: bool = False,
) -> str:
    src = None
    try:
        src = get_source_code_lineno()

        if explanation:
            _log(explanation)

        dels = set()
        for start, count in deletions.items():
            start = int(start)
            dels.update(range(start, start + count))
        modifications = {int(k): v for k, v in modifications.items()}
        insertions = {int(k): v for k, v in insertions.items()}
        res = []
        if 0 in insertions:
            res += insertions[0]
        for line, code in src:
            if line in modifications:
                res.append(modifications[line])
            elif line in insertions:
                res += insertions[line]
                if line not in dels:
                    res.append(code)
            else:
                if line not in dels:
                    res.append(code)
        while line + 1 in insertions:
            res += insertions[line + 1]
            line += 1
        path = os.path.join(get_cwd(), "features.py")
        with open(path, "w") as f:
            f.write("\n".join(res))
        validate_syntax(path)
        run_command(f"black {path}")
        return _success("Feature definitions updated without syntax issues")
    except SyntaxError as se:
        if revert_on_error:
            with open(path, "w") as f:
                f.write("\n".join([x[1] for x in src]))
            return _err(
                f"{se}, the change has been reverted, try again!", prefix="Syntax error"
            )
        else:
            return _err(se, prefix="Syntax error")
    except Exception as e:
        return _err(e)
    finally:
        _code()



from typing import Any, Dict, Optional, Type, Union

from llama_index.core.agent import FunctionCallingAgentWorker
from llama_index.core.agent.runner.base import AgentRunner
from llama_index.core.llms import ChatMessage, MessageRole
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.prompts import ChatPromptTemplate
from pydantic import BaseModel

from ..agent.base import AgentBase, _IntegratedAgent, invoke_agent, make_agent
from ..factories.llm import make_llm_model


@make_llm_model.register(FunctionCallingLLM)
def make_llama_index_model(obj: FunctionCallingLLM) -> FunctionCallingLLM:
    return obj


@invoke_agent.register(FunctionCallingLLM)
def _invoke_llama(
    llm: FunctionCallingLLM,
    client: AgentBase,
    input: Dict[str, Any],
    output_schema: Union[Type[str], Type[BaseModel]] = str,
    **kwargs: Any,
) -> str:
    runner = make_llama_index_agent_runner(
        llm, client, output_schema=output_schema, **kwargs
    )
    client.logger.debug(
        "Invoking LlamaIndex agent with %s",
        input,
    )
    message = input.get("message")
    res = runner.chat(message, chat_history=input.get("chat_history"))
    if issubclass(output_schema, BaseModel):
        messages = ChatPromptTemplate(
            message_templates=[
                ChatMessage(role=MessageRole.USER, content=message),
                ChatMessage(role=MessageRole.ASSISTANT, content=str(res)),
                ChatMessage(
                    role=MessageRole.USER, content="convert the response to structured"
                ),
            ]
        )
        res = llm.structured_predict(output_cls=output_schema, prompt=messages)
        return res
    else:
        return str(res)


@make_agent.register(FunctionCallingLLM)
def make_llama_index_agent_runner(
    llm: FunctionCallingLLM,
    client: AgentBase,
    output_schema: Union[Type[str], Type[BaseModel]] = str,
    initial_input: Optional[Dict[str, Any]] = None,
    **runner_kwargs: Any,
) -> AgentRunner:
    agent = _LlamaIndexAgent(client, llm, input=initial_input)
    return agent.make_agent_runner(**runner_kwargs)


class _LlamaIndexAgent(_IntegratedAgent):
    def make_agent_runner(self, **kwargs):
        agent_worker = _TectonFunctionCallingAgentWorker.from_tools(
            self.tools,
            llm=self.llm,
            verbose=kwargs.get("verbose", False),
            allow_parallel_tool_calls=True,
            tecton_agent=self,
        )
        agent = agent_worker.as_agent(**kwargs)
        return agent

    def _make_tool(self, name):
        from llama_index.core.tools import FunctionTool

        def f(**kwargs):
            pass

        model = self._get_tool_schema(name)
        _tool = FunctionTool.from_defaults(
            fn=f, fn_schema=model, name=name, description=model.__doc__
        )
        _tool._fn = lambda **kwargs: self.client.invoke_tool(name, kwargs)
        return _tool


class _TectonFunctionCallingAgentWorker(FunctionCallingAgentWorker):
    def __init__(self, *args, tecton_agent: _LlamaIndexAgent, **kwargs):
        super().__init__(*args, **kwargs)
        self._tecton_agent = tecton_agent

    def _make_messages(self, question, history):
        history = history or []
        history = self._add_sys_prompt(history)
        return history + [("human", question)]

    def get_all_messages(self, task):
        sys_prompt = self._tecton_agent.client.invoke_system_prompt()
        self.prefix_messages = [
            ChatMessage(role=MessageRole.SYSTEM, content=sys_prompt)
        ]
        return super().get_all_messages(task)



from contextlib import contextmanager
from functools import singledispatch
from logging import Logger
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from langchain.agents import AgentExecutor, create_tool_calling_agent
from langchain.agents.format_scratchpad.tools import format_to_tool_messages
from langchain.agents.output_parsers.tools import ToolsAgentOutputParser
from langchain.chains.base import Chain
from langchain_community.callbacks import get_openai_callback
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import RunnablePassthrough
from langchain_core.tools import StructuredTool
from langchain_core.vectorstores import VectorStore
from pydantic import BaseModel

from ..agent.base import (
    AgentBase,
    _IntegratedAgent,
    invoke_agent,
    invoke_emitters,
    make_agent,
)
from ..factories.llm import (
    invoke_llm_model,
    make_llm,
    make_llm_model,
    make_llm_model_conf,
)
from ..factories.vector_db import make_vector_db_conf, make_vector_db_instance
from ..utils.config_wrapper import as_config
from ..utils.hashing import to_uuid
from ..utils.log import report_cost

_LLM_PROVIDERS: dict[str, Callable] = {}


@make_llm_model.register(BaseChatModel)
def make_langchain_model(obj: BaseChatModel) -> BaseChatModel:
    return obj


@invoke_llm_model.register(BaseChatModel)
def invoke_langchain_model(
    llm: BaseChatModel,
    prompt: Union[str, List[Tuple[str, str]]],
    output_schema: Union[Type[BaseModel], Type[str]] = str,
) -> Any:
    with track_llm_cost():
        if output_schema is not str:
            llm = llm.with_structured_output(output_schema)
            return llm.invoke(prompt)
        return llm.invoke(prompt).content


@make_llm_model.register(Embeddings)
def make_langchain_embedding_model(obj: Embeddings) -> Embeddings:
    return obj


@make_llm_model.register(dict)
def make_langchain_model_from_dict(
    model_conf: dict,
) -> Union[BaseChatModel, Embeddings]:
    model, conf = _parse_dict(model_conf)
    return model(**conf)


@make_llm_model_conf.register(dict)
def make_langchain_model_conf_from_dict(model_conf: dict) -> dict:
    model, conf = _parse_dict(model_conf)
    return as_config(model)(**conf)


@contextmanager
def track_llm_cost():
    with get_openai_callback() as cb:
        try:
            yield cb
        finally:
            report_cost(cb.total_cost, cb.successful_requests)


def langchain_llm_provider(name: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        _LLM_PROVIDERS[name] = func
        return func

    return decorator


@langchain_llm_provider("openai")
def _make_openai_llm(model_name: str, conf: dict) -> Tuple[Type, Dict]:
    if "embedding" in model_name:
        from langchain_openai import OpenAIEmbeddings

        return OpenAIEmbeddings, conf
    else:
        from langchain_openai import ChatOpenAI

        return ChatOpenAI, conf


@langchain_llm_provider("azure_openai")
def _make_azure_openai_llm(model_name: str, conf: dict) -> Tuple[Type, Dict]:
    if "embedding" in model_name:
        from langchain_openai import AzureOpenAIEmbeddings

        return AzureOpenAIEmbeddings, conf
    else:
        from langchain_openai import AzureChatOpenAI

        return AzureChatOpenAI, conf


@langchain_llm_provider("anthropic")
def _make_anthropic_llm(model_name: str, conf: dict) -> Tuple[Type, Dict]:
    from langchain_anthropic import ChatAnthropic

    return ChatAnthropic, conf


@langchain_llm_provider("bedrock")
def _make_bedrock_llm(model_name: str, conf: dict) -> Tuple[Type, Dict]:
    if "embed" in model_name:
        from langchain_aws import BedrockEmbeddings

        conf["model_id"] = conf.pop("model")
        return BedrockEmbeddings, conf
    else:
        from langchain_aws import ChatBedrock

        return ChatBedrock, conf


@langchain_llm_provider("google")
def _make_google_llm(model_name: str, conf: dict) -> Tuple[Type, Dict]:
    if "embed" in model_name:
        from langchain_google_genai import GoogleGenerativeAIEmbeddings

        return GoogleGenerativeAIEmbeddings, conf
    else:
        from langchain_google_genai import ChatGoogleGenerativeAI

        return ChatGoogleGenerativeAI, conf


def _parse_dict(data: Dict[str, Any]) -> Tuple[Type, Dict]:
    conf = dict(data)
    parts = conf.pop("model").split("/", 1)
    provider, model_name = parts[0], parts[1]
    conf["model"] = model_name
    if provider in _LLM_PROVIDERS:
        return _LLM_PROVIDERS[provider](model_name, conf)
    raise NotImplementedError(f"{provider} not supported")


@make_vector_db_instance.register(VectorStore)
def make_langchain_vector_db_instance(obj: VectorStore) -> VectorStore:
    return obj


@make_vector_db_instance.register(dict)
def make_langchain_vector_db_instance_from_dict(
    vdb_conf: dict,
) -> VectorStore:
    vdb, conf = _parse_vdb_dict(vdb_conf)
    return vdb(**conf)


@make_vector_db_conf.register(dict)
def make_langchain_vector_db_conf_from_dict(vdb_conf: dict) -> dict:
    vdb, conf = _parse_vdb_dict(vdb_conf, as_conf=True)
    return as_config(vdb)(**conf)


def _parse_vdb_dict(data: Dict[str, Any], as_conf: bool = False) -> Tuple[Type, Dict]:
    conf = dict(data)
    provider = conf.pop("provider")
    embedding = make_llm(conf.pop("embedding"), conf=as_conf)
    conf["embedding"] = embedding
    if provider == "memory":
        from langchain_community.vectorstores.inmemory import InMemoryVectorStore

        return InMemoryVectorStore, conf
    if provider == "lancedb":
        from langchain_community.vectorstores.lancedb import LanceDB

        return LanceDB, conf
    if provider == "pinecone":
        from langchain_pinecone import PineconeVectorStore

        return PineconeVectorStore, conf
    raise NotImplementedError(f"{provider} not supported")


@invoke_agent.register(BaseChatModel)
def _invoke_langchain(
    llm: BaseChatModel,
    client: AgentBase,
    input: Dict[str, Any],
    output_schema: Union[Type[str], Type[BaseModel]] = str,
    **kwargs: Any,
) -> Any:
    with track_llm_cost():
        callbacks = list(kwargs.pop("callbacks", []))
        cb = _ExecutionCallback(client.logger, client)
        callbacks.append(cb)
        executor = make_langchain_agent_executor(
            llm,
            client,
            output_schema,
            initial_input=input,
            **kwargs,
        )
        chat_history = input.get("chat_history")
        client.logger.debug("Invoking LangChain agent with input: %s", input)
        _input = {"input": input["message"]}
        if chat_history:
            _input["chat_history"] = chat_history
        try:
            return executor.invoke(_input, {"callbacks": callbacks})
        except _StructuredOutput as e:
            return e.output


@invoke_emitters.register(BaseChatModel)
def _invoke_langchain_emitters(
    llm: BaseChatModel,
    client: AgentBase,
    input: Dict[str, Any],
    **kwargs: Any,
) -> None:
    agent = _LangChainAgent(client, llm, is_emitter=True, input=input)
    agent.run_emitters(input, **kwargs)


@make_agent.register(BaseChatModel)
def make_langchain_agent_executor(
    llm: BaseChatModel,
    client: AgentBase,
    output_schema: Union[Type[str], Type[BaseModel]] = str,
    initial_input: Optional[Dict[str, Any]] = None,
    **executor_kwargs: Any,
) -> Chain:
    agent = _LangChainAgent(
        client,
        llm,
        is_emitter=False,
        input=initial_input,
    )
    return agent.make_executor(output_schema=output_schema, **executor_kwargs)


@singledispatch
def unified_langchain_vector_search(
    vdb: VectorStore,
    query: str,
    top_k: int,
    search_type: str = "similarity",
    **params: Any,
) -> List[Document]:
    return vdb.search(query, k=top_k, search_type=search_type, **params)


def langchain_vector_search(
    vdb: VectorStore,
    query: str,
    top_k: int,
    search_type: str = "similarity",
    **params: Any,
) -> List[Document]:
    try:
        from .lancedb import _lancedb_vector_search  # noqa
    except ImportError:
        pass

    return unified_langchain_vector_search(
        vdb, query=query, top_k=top_k, search_type=search_type, **params
    )


def langchain_vector_ingest(
    vdb: VectorStore,
    texts: List[str],
    metadatas: Optional[List[Dict[str, Any]]] = None,
    ids: Optional[List[str]] = None,
) -> None:
    if ids is None:
        ids = [to_uuid(text) for text in texts]
    if metadatas is None:
        metadatas = [{} for _ in texts]
    vdb.add_texts(texts=texts, metadatas=metadatas, ids=ids)


class _LangChainAgent(_IntegratedAgent):
    def make_executor(self, output_schema: Any, **kwargs: Any) -> Chain:
        tools = self.tools
        prompt = self._make_prompt_template(tools)
        if len(tools) > 0:
            if output_schema is str:

                def parse(message: Dict[str, Any]) -> Any:
                    output = message["output"]
                    if isinstance(output, list):
                        output = output[0]["text"]
                    return output

                return (
                    _make_lc_agent_executor(
                        llm=self.llm, tools=tools, prompt=prompt, **kwargs
                    )
                    | parse
                )

            else:

                def _final_structured_output(**kwargs):
                    raise _StructuredOutput(kwargs)

                _final_tool = StructuredTool.from_function(
                    _final_structured_output,
                    description="""
Final output tool for structured output.
This tool must be used once and only once in the end to generate the final structured output.

"""
                    + (output_schema.__doc__ or ""),
                    args_schema=output_schema,
                    infer_schema=False,
                )
                return _make_lc_agent_executor(
                    llm=self.llm,
                    tools=tools + [_final_tool],
                    prompt=prompt,
                    bind_kwargs={"tool_choice": "any"},
                    **kwargs,
                )
        else:
            if issubclass(output_schema, BaseModel):
                llm = self.llm.with_structured_output(output_schema)

                return prompt | llm

            def parse(message: AIMessage) -> str:
                return message.content

            return prompt | self.llm | parse

    def run_emitters(self, input: Dict[str, Any], **kwargs: Any) -> None:
        message = input["message"]
        chat_history = input.get("chat_history")
        tools = self.tools
        if len(tools) == 0:
            return
        prompt = self._make_prompt_template(tools)
        metastore = self.client.metastore[self.meta_key]
        kwargs = kwargs.copy()
        kwargs["max_iterations"] = 3
        emitters_no_history = [e for e in tools if not metastore[e.name]["use_history"]]
        emitters_with_history = [e for e in tools if metastore[e.name]["use_history"]]
        if len(emitters_no_history) > 0:
            agent = create_tool_calling_agent(
                self.llm, emitters_no_history, prompt=prompt
            )
            executor = AgentExecutor(
                agent=agent, tools=emitters_no_history, **kwargs, verbose=True
            )
            executor.invoke({"input": message})
        if len(emitters_with_history) > 0:
            agent = create_tool_calling_agent(
                self.llm, emitters_with_history, prompt=prompt
            )
            executor = AgentExecutor(
                agent=agent, tools=emitters_with_history, **kwargs, verbose=True
            )
            executor.invoke({"input": message, "chat_history": chat_history})

    def _make_prompt_template(self, tools: List[Any]):
        templates = [
            MessagesPlaceholder("system_prompt"),
            MessagesPlaceholder("chat_history", optional=True),
            ("human", "{input}"),
        ]
        if len(tools) > 0:
            templates.append(MessagesPlaceholder("agent_scratchpad"))
        prompt = ChatPromptTemplate.from_messages(templates)
        prompt = prompt.partial(
            system_prompt=lambda: [("system", self.client.invoke_system_prompt())]
        )
        return prompt

    def _make_messages(self, question, history):
        history = history or []
        history = self._add_sys_prompt(history)
        return history + [("human", question)]

    def _make_tool(self, name):
        from langchain_core.tools import StructuredTool

        model = self._get_tool_schema(name)

        def f(**kwargs):
            pass

        _tool = StructuredTool.from_function(
            name=name,
            func=f,
            args_schema=model,
            infer_schema=False,
            description=model.__doc__,
        )
        if self.meta_key == "tools":
            _tool.func = lambda **kwargs: self.client.invoke_tool(name, kwargs)
        else:
            _tool.func = lambda **kwargs: self.client.invoke_emitter(name, kwargs)
        return _tool


class _ExecutionCallback(BaseCallbackHandler):
    def __init__(self, logger: Logger, client: AgentBase):
        self.metastore = client.metastore
        self.logger = logger

    def on_chat_model_start(
        self, serialized: Dict[str, Any], messages: List[List[BaseMessage]], **kwargs
    ) -> None:
        self.logger.debug("Chat model started", extra={"flow_event": {"type": "llm"}})

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        name = serialized.get("name")
        extra = {"flow_event": {"type": "tool", "value": name}}
        tool = self.metastore.get("tools", {}).get(name, {})
        extra["flow_event"]["subtype"] = tool.get("subtype", "")
        if tool.get("subtype") == "search":
            extra["flow_event"]["knowledge"] = tool.get("source_names", [])
        else:
            extra["flow_event"]["features"] = tool.get("source_names", [])
        self.logger.debug(f"Tool {name} started", extra=extra)


class _StructuredOutput(Exception):
    def __init__(self, output: Any):
        super().__init__()
        self.output = output


def _make_lc_agent_executor(
    llm: Any,
    tools: List[Any],
    prompt: Any,
    bind_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> AgentExecutor:
    # a copy from create_tool_calling_agent because it doesn't have bind_kwargs
    llm_with_tools = llm.bind_tools(tools, **(bind_kwargs or {}))

    agent = (
        RunnablePassthrough.assign(
            agent_scratchpad=lambda x: format_to_tool_messages(x["intermediate_steps"])
        )
        | prompt
        | llm_with_tools
        | ToolsAgentOutputParser()
    )

    ae = AgentExecutor(agent=agent, tools=tools, **kwargs)
    ae.agent.stream_runnable = False
    return ae



import datetime
import json
import os

from constants import FILE_DIR
from utils import get_cwd, is_in_feature_repo, get_tecton_account_info

import logfire

# configure logfire
LOGFIRE_API_KEY = os.getenv("LOGFIRE_API_KEY")
if LOGFIRE_API_KEY:
    logfire.configure(token=LOGFIRE_API_KEY)
    logfire.instrument_openai()
    print("Tracing to https://logfire.pydantic.dev/")
else:
    print("LOGFIRE_API_KEY is not set, so logfire will not be used")

def sys_prompt() -> str:
    context = {
        "Current directory": get_cwd(),
        "In feature repo": is_in_feature_repo(),
        "Current Tecton account info": get_tecton_account_info(),
        "Current time": datetime.datetime.now().isoformat(),
    }
    with open(os.path.join(FILE_DIR, "data", "gotchas.md"), "r") as f:
        gotchas = f.read()
    with open(os.path.join(FILE_DIR, "data", "sys_prompt.md"), "r") as f:
        return f.read().format(context=json.dumps(context, indent=4), gotchas=gotchas)



_TECTON_CHECKPOINT_ATTR = "__tecton_checkpoint__"
_TECTON_MOCK_OBJECT_ATTR = "_tecton_mock_object_attr"
_DEV_MODE_AGENT_KEY = "dev_mode_agent"
_INPUT_JSON_FIELD = "input_json"
_OUTPUT_JSON_FIELD = "output_json"
UUID_ATTR = "__uuid__"
DEFAULT_TIMEOUT = "5s"
TECTON_DEV_MODE_KEY = "TECTON_GEN_AI_DEV_MODE"



import base64
import datetime
import os
from tempfile import gettempdir
from typing import Dict, List, Annotated

import pandas as pd
import pyarrow as pa
from pyarrow.fs import FileType
from pytest import raises
from tecton.types import (
    Array,
    Bool,
    Field,
    Float64,
    Int32,
    Int64,
    Map,
    String,
    Timestamp,
)
from tecton_core import conf as tecton_conf

from melvin.tecton_gen_ai.tecton_utils.convert import (
    to_tecton_type,
    tecton_type_to_python_annotation,
)

try:
    from tecton_proto.materialization.params_pb2 import MaterializationTaskParams
except ImportError:
    from tecton_proto.materialization.params__client_pb2 import (
        MaterializationTaskParams,
    )

from melvin.tecton_gen_ai.tecton_utils._tecton_utils import (
    _CACHE_PREFIX,
    featurize_async,
    get_cache_base_path,
    get_df_schema,
    run_async,
    run_async_jobs,
    set_conf,
)


def test_tecton_type_to_python_type():
    def _assert(tp, t):
        assert tecton_type_to_python_annotation(tp) == t

    _assert(String, str)
    _assert(Int64, int)
    _assert(Float64, float)
    _assert(Bool, bool)
    _assert(Array(Int64), List[int])
    _assert(Array(Array(Int64)), List[List[int]])
    _assert(Map(String, Int64), Dict[str, int])
    _assert(Timestamp, datetime.datetime)

    with raises(ValueError):
        _assert(None, None)


def test_type_annotation_to_tecton_type():
    def _assert(tp, t):
        assert str(to_tecton_type(tp)) == str(t)

    _assert(str, String)
    _assert(int, Int64)
    _assert(float, Float64)
    _assert(bool, Bool)
    _assert(Annotated[bool, "abc"], Bool)
    _assert(List[int], Array(Int64))
    _assert(List[list[int]], Array(Array(Int64)))
    _assert(Dict[str, int], Map(String, Int64))
    _assert(list[int], Array(Int64))
    _assert(dict[str, int], Map(String, Int64))
    _assert(pd.Timestamp, Timestamp)
    _assert(datetime.datetime, Timestamp)

    _assert(pa.int32(), Int32)
    _assert(pa.int64(), Int64)
    _assert(pa.float32(), Float64)
    _assert(pa.float64(), Float64)
    _assert(pa.string(), String)
    _assert(pa.timestamp("s"), Timestamp)
    _assert(pa.list_(pa.int32()), Array(Int32))

    with raises(ValueError):
        _assert(pa.map_(pa.int32(), pa.int32()), Map(Int32, Int32))

    with raises(ValueError):
        _assert(None, None)


def test_get_df_schema():
    df = pd.DataFrame(
        [
            dict(
                a=1,
                b=1.0,
                c="a",
                d=True,
                e=pd.Timestamp("2021-01-01"),
                f=[1, 2, 3],
            ),
        ]
    )
    schema = get_df_schema(df)
    assert schema == [
        Field("a", Int64),
        Field("b", Float64),
        Field("c", String),
        Field("d", Bool),
        Field("e", Timestamp),
        Field("f", Array(Int64)),
    ]

    with raises(ValueError):
        df = pd.DataFrame([dict(a={})])
        get_df_schema(df)


def test_get_cache_base_path(tmp_path):
    # test service mode
    temp_dir = str(tmp_path)
    params = MaterializationTaskParams()
    params.offline_store_path = os.path.join(temp_dir, "test", "randomid", "data")
    data = params.SerializeToString()
    data_str = base64.standard_b64encode(data).decode("utf-8")
    os.environ["MATERIALIZATION_TASK_PARAMS"] = data_str
    fs, p = get_cache_base_path()
    assert fs is not None
    assert p == os.path.join(temp_dir, "test", _CACHE_PREFIX)
    assert fs.get_file_info(p).type == FileType.Directory
    with raises(ValueError):
        # The offline_store_path format is invalid
        params.offline_store_path = os.path.join(temp_dir, "test", "randomid")
        data = params.SerializeToString()
        data_str = base64.standard_b64encode(data).decode("utf-8")
        os.environ["MATERIALIZATION_TASK_PARAMS"] = data_str
        fs, p = get_cache_base_path()

    # test dev mode
    temp_dir = gettempdir()
    del os.environ["MATERIALIZATION_TASK_PARAMS"]
    fs, p = get_cache_base_path()
    assert fs is not None
    assert p == os.path.join(temp_dir, _CACHE_PREFIX)
    assert fs.get_file_info(p).type == FileType.Directory


def test_run_async():
    async def _coro():
        return 1

    assert run_async(_coro()) == 1

    async def _coro():
        raise ValueError("error")

    with raises(ValueError):
        run_async(_coro())


def test_run_async_jobs():
    async def _coro(n):
        if n < 2:
            return n
        raise ValueError("error")

    assert run_async_jobs([_coro(0), _coro(1)], concurrency=1) == [0, 1]

    with raises(ValueError):
        run_async_jobs([_coro(0), _coro(2)], concurrency=1)


def test_featurize_async():
    async def _f1(n):
        return n + 10

    async def _f2(batch):
        return [n + 20 for n in batch]

    df = pd.DataFrame(dict(a=[1, 2, 1]))
    res = featurize_async(df, "a", "b", _f1, concurrency=2, dedup=False)
    assert res.b.tolist() == [11, 12, 11]

    df = pd.DataFrame(dict(a=[1, 2, 1]))
    res = featurize_async(
        df, "a", "b", _f2, concurrency=2, dedup=False, mini_batch_size=2
    )
    assert res.b.tolist() == [21, 22, 21]


def test_set_conf():
    tecton_conf.set("TECTON_FORCE_FUNCTION_SERIALIZATION", "false")
    assert tecton_conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION") == "false"
    with set_conf({"TECTON_FORCE_FUNCTION_SERIALIZATION": "true"}):
        assert tecton_conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION") == "true"
    assert tecton_conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION") == "false"

    tecton_conf.set("TECTON_FORCE_FUNCTION_SERIALIZATION", None)
    with set_conf({"TECTON_FORCE_FUNCTION_SERIALIZATION": "true"}):
        assert tecton_conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION") == "true"
    # assert tecton_conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION") is None



from typing import Any, Callable, Dict, List, Optional, TypeVar

from pydantic import BaseModel
from tecton import FeatureView, RequestSource
from typing_extensions import ParamSpec
import inspect
from ._internal import FuncWrapper
from ..utils._serialization import to_openai_function
from ..utils.config_utils import Configs

_METASTORE: Dict[str, Any] = {}

T = TypeVar("T")
P = ParamSpec("P")


def _get_from_metastore(obj: Any, register: Any = None) -> Dict[str, Any]:
    key = id(obj)
    if key not in _METASTORE:
        if register is not None:
            register()
            return _get_from_metastore(obj)
        raise ValueError(
            f"{obj} not found in metastore, did you forget to decorate it?"
        )
    return _METASTORE[key]


def _set_metastore(key: Any, value: Any) -> None:
    _METASTORE[id(key)] = value


def _get_source_names(sources: Optional[List[FeatureView]]) -> List[str]:
    if sources is None:
        return []
    return [source.name for source in sources if not isinstance(source, RequestSource)]


def prompt(
    func: Optional[Callable[P, T]] = None,
    *,
    name: Optional[str] = None,
    sources: Optional[List[FeatureView]] = None,
    timeout: Any = None,
    **prompt_kwargs: Any,
) -> Callable[P, T]:
    """
    Decorator for creating a prompt.

    Args:

        func: The function to decorate.
        name: The name of the prompt.
        sources: The sources of the prompt.
        timeout: The timeout expression of the tool, defaults to None (use the default timeout).
            The expression must be accepted by `pandas.to_timedelta`.
        **prompt_kwargs: Other keyword arguments for Tecton's prompt.

    Examples:

        ```python
        # You don't need to use the decorator if you don't need to change any parameter

        def simple_prompt() -> str:
            return "Hello, world!"

        # With parameters but not a feature

        def prompt_with_params(zip:str) -> str:
            return "Hello, you zip code " + zip

        # When you need to customize the properties of the prompt
        from tecton_gen_ai.api import prompt

        @prompt(name="simple_prompt_2", timeout="10s")
        def simple_prompt() -> str:
            return "Hello, world!"

        # With a feature dependency
        from tecton_gen_ai.testing import make_local_batch_feature_view

        my_bfv = make_local_batch_feature_view(
            "my_data",
            {"user_id":"u1", "name":"Bob"},
            ["user_id"],
            description="User information",
        )

        @prompt(sources=[my_bfv])
        def sys_prompt(zip:str, my_data) -> str:  # the parameter name of the bfv must match the name of the bfv
            return "Hello, " + my_data["name"] + " with zip code " + zip

        # Serve the prompt
        from tecton_gen_ai.api import Agent

        agent = Agent(
            name="my_service",
            prompt=sys_prompt,
        )

        # Test locally
        with agent.set_context({"zip": "10065", "user_id": "u1"}):
            res = agent.invoke_prompt()
        ```
    """

    kwargs = prompt_kwargs.copy()
    description = kwargs.pop("description", "Prompt")
    return _internal_tool(
        func=func,
        name=name,
        sources=sources,
        description=description,
        type="prompt",
        subtype="prompt",
        timeout=timeout,
        **kwargs,
    )


# TODO: temporarily disable this decorator because prompt FCO doesn't support secrets and resources
def _prompt_not_in_use(
    func: Optional[Callable[P, T]] = None,
    name: Optional[str] = None,
    sources: Optional[List[FeatureView]] = None,
    timeout: Any = None,
    **prompt_kwargs: Any,
) -> Callable[P, T]:  # pragma: no cover
    """
    Decorator for creating a prompt.

    Args:

        func: The function to decorate.
        name: The name of the prompt.
        sources: The sources of the prompt.
        timeout: The timeout expression of the tool, defaults to None (use the default timeout).
            The expression must be accepted by `pandas.to_timedelta`.
        **prompt_kwargs: Other keyword arguments for Tecton's prompt.

    Examples:

        ```python
        # You don't need to use the decorator if you don't need to change any parameter

        def simple_prompt() -> str:
            return "Hello, world!"

        # With parameters but not a feature

        def prompt_with_params(zip:str) -> str:
            return "Hello, you zip code " + zip

        # When you need to customize the properties of the prompt
        from tecton_gen_ai.api import prompt

        @prompt(name="simple_prompt_2", timeout="10s")
        def simple_prompt() -> str:
            return "Hello, world!"

        # With a feature dependency
        from tecton_gen_ai.testing import make_local_batch_feature_view

        my_bfv = make_local_batch_feature_view(
            "my_data",
            {"user_id":"u1", "name":"Bob"},
            ["user_id"],
            description="User information",
        )

        @prompt(sources=[my_bfv])
        def sys_prompt(zip:str, my_data) -> str:  # the parameter name of the bfv must match the name of the bfv
            return "Hello, " + my_data["name"] + " with zip code " + zip

        # Serve the prompt
        from tecton_gen_ai.api import Agent

        agent = Agent(
            name="my_service",
            prompt=sys_prompt,
        )

        # Test locally
        with agent.set_context({"zip": "10065", "user_id": "u1"}):
            res = agent.invoke_prompt()
        ```
    """

    def wrapper(
        _func: Callable[P, T],
        _name: Optional[str],
        _features: Optional[List[FeatureView]],
    ) -> Callable[P, T]:
        if _name is None:
            _name = _func.__name__
        wrapper = FuncWrapper(_name, _func, _features)
        fv = wrapper.make_prompt(**prompt_kwargs)
        _set_metastore(
            _func,
            {
                "name": _name,
                "fv": fv,
                "type": "prompt",
                "llm_args": wrapper.llm_args,
                "entity_args": wrapper.entity_args,
                "feature_args": wrapper.feature_args,
                "source_names": _get_source_names(_features),
                "timeout_sec": Configs.get_default().get_timeout_sec(timeout),
            },
        )
        return _func

    if func is None:
        return lambda _func: wrapper(_func, name, sources)
    return wrapper(func, name, sources)


def tool(
    func: Optional[Callable[P, T]] = None,
    *,
    name: Optional[str] = None,
    sources: Optional[List[FeatureView]] = None,
    description: Optional[str] = None,
    use_when: Optional[Callable] = None,
    timeout: Any = None,
    **rtfv_kwargs: Any,
) -> Callable[P, T]:
    """
    Decorator for creating a tool. The tool can be used by different LLMs and
    frameworks such as Langchain and LlamaIndex.

    Args:

        func: The function to decorate.
        name: The name of the tool.
        sources: The sources of the tool.
        description: The description of the tool, defaults to None (the docstring of the function).
        use_when: The condition function to use the tool, defaults to None. If set, it must be
            a defined function. And the input should be a string message and the list of chat history.
        timeout: The timeout expression of the tool, defaults to None (use the default timeout).
            The expression must be acceptable by `pandas.to_timedelta`.
        **rtfv_kwargs: Other keyword arguments for Tecton's realtime feature view.

    Note:

        The definition of tool is different from prompt:

        - The name of the tool should represent what the function does.
        - The docstring of the tool is required and it should contain the description and instructions of the tool.
        - If a feature view is a dependency, then the entity ids must be defined in the function signature.
        - All the arguments excluding feature view arguments must have type annotations.
        - The return type annotation is required.

    Examples:

        ```python
        # In many cases you don't need to use the decorator, `@tool` can be omitted
        def get_months_in_a_year() -> int:
            '''The total number months of a year'''
            return 12

        # With parameters but not a feature
        def add(a: int, b: int) -> int:
            '''Add two numbers

            Args:
                a (int): The first number
                b (int): The second number

            Returns:
                int: The sum of the two numbers
            '''
            return a + b

        # Serve the tool
        from tecton_gen_ai.api import Agent

        agent = Agent(name="my_service", tools=[get_months_in_a_year, add])
        agent.invoke_tool("get_months_in_a_year")
        agent.invoke_tool("add", dict(a=1, b=2))

        # With a feature dependency
        from tecton_gen_ai.api import tool
        from tecton_gen_ai.testing import make_local_batch_feature_view

        my_bfv = make_local_batch_feature_view(
            "my_bfv",
            {"user_id":"u1", "name":"Bob"},
            ["user_id"],
            description="User information",
        )

        @tool(sources=[my_bfv], timeout="10s")
        def get_name(prefix:str, user_id:str, my_bfv) -> str:
            '''Get the name of a user with a prefix

            Args:

                prefix (str): The prefix of the name
                user_id (str): The user ID

            Returns:

                str: The name of the user with the prefix
            '''
            return prefix + " " + my_bfv["name"]

        # Serve the tool
        from tecton_gen_ai.api import Agent

        agent = Agent(name="my_service", tools=[get_name])

        with agent.set_context({"user_id": "u1", "prefix": "Hello"}):
            res = agent.invoke_tool("get_name")
        ```
    """

    return _internal_tool(
        func=func,
        name=name,
        sources=sources,
        description=description,
        use_when=use_when,
        subtype="tool",
        timeout=timeout,
        **rtfv_kwargs,
    )


def emitter(
    func: Optional[Callable[P, T]] = None,
    *,
    name: Optional[str] = None,
    sources: Optional[List[FeatureView]] = None,
    description: Optional[str] = None,
    use_when: Optional[Callable] = None,
    use_history: bool = False,
    timeout: Any = None,
    **rtfv_kwargs: Any,
) -> Callable[P, T]:
    # TODO: is it too heavy to let each emitter be a feature view?
    # TODO: Add doc string
    return _internal_emitter(
        func=func,
        name=name,
        sources=sources,
        description=description,
        use_when=use_when,
        use_history=use_history,
        timeout=timeout,
        **rtfv_kwargs,
    )


def _internal_emitter(
    *,
    func: Optional[Callable[P, T]] = None,
    name: Optional[str] = None,
    sources: Optional[List[FeatureView]] = None,
    description: Optional[str] = None,
    use_history: bool = False,
    timeout: Any = None,
    input_schema: Optional[BaseModel] = None,
    **rtfv_kwargs: Any,
) -> Callable[P, T]:
    # assert return type is None
    return _internal_tool(
        func=func,
        name=name,
        sources=sources,
        description=description,
        type="emitter",
        subtype="emitter",
        timeout=timeout,
        input_schema=input_schema,
        extra_metadata={"use_history": use_history},
        **rtfv_kwargs,
    )


def _internal_tool(
    func: Optional[Callable[P, T]] = None,
    name: Optional[str] = None,
    sources: Optional[List[FeatureView]] = None,
    description: Optional[str] = None,
    type: str = "tool",
    subtype: str = "tool",
    source_names: Optional[List[str]] = None,
    extra_metadata: Optional[Dict[str, Any]] = None,
    timeout: Any = None,
    input_schema: Optional[BaseModel] = None,
    use_when: Optional[Callable] = None,
    **rtfv_kwargs: Any,
) -> Callable[P, T]:
    def wrapper(
        _func: Callable[P, T],
        _name: Optional[str],
        _features: Optional[List[FeatureView]],
    ) -> Callable[P, T]:
        if _name is None:
            _name = _func.__name__
        wrapper = FuncWrapper(
            _name,
            _func,
            _features,
            assert_entity_defined=(type == "tool"),
            input_schema=input_schema,
        )
        fv = wrapper.make_feature_view(**rtfv_kwargs)
        # assert use_when is not a lambda function
        if use_when is not None:
            if (
                inspect.isfunction(use_when)
                and use_when.__name__ != "<lambda>"
            ):
                pass
            else:
                raise ValueError("use_when must be a defined function")
        excluded_args = wrapper.feature_args
        metadata = {
            "name": _name,
            "fv": fv,
            "use_when": use_when,
            "type": type,
            "subtype": subtype,
            "llm_args": wrapper.llm_args,
            "entity_args": wrapper.entity_args,
            "feature_args": wrapper.feature_args,
            "function": to_openai_function(
                input_schema or _func,
                name=_name,
                exclude_args=excluded_args,
                description=description or _func.__doc__,
            ),
            "source_names": source_names or _get_source_names(_features),
            "timeout_sec": Configs.get_default().get_timeout_sec(timeout),
        }
        if extra_metadata is not None:
            metadata.update(extra_metadata)
        _set_metastore(_func, metadata)
        return _func

    if func is None:
        return lambda _func: wrapper(_func, name, sources)
    return wrapper(func, name, sources)



from datetime import datetime
import streamlit as st
from agents import build_agent
from constants import MELVIN_PATH
from lib import parse_args
from utils import (
    _code,
    display,
    initialize_tecton_account_info_if_needed,
    get_tecton_account_info,
    load_object,
    set_code_container,
)
from editors.feature_view_editor_agent import display_fv_editor

from melvin.tecton_gen_ai.api import Configs
from melvin.tecton_gen_ai.testing import set_dev_mode
from melvin.tecton_gen_ai.testing.log_utils import UILogHandler, set_ui_logger
from melvin.tecton_gen_ai.utils.log import cost_monitor

# We need to do this hack because set_dev_mode below actually overrides some env vars that get_current_tecton_account_info depends on
initialize_tecton_account_info_if_needed()
print(get_tecton_account_info())

set_dev_mode()

st.set_page_config(page_title="Tecton Copilot", layout="wide")

def _print_record(record) -> bool:
    flag = getattr(record, "copilot_flag", None)
    msg = record.msg
    if flag == "log":
        st.markdown(msg)
    elif flag == "log_with_details":
        print(record)
        st.markdown(msg)
        with st.popover("Query Details"):
            st.markdown(record.details)
    elif flag == "image":
        image_html = (
            f'<img src="data:image/png;base64,{record.image_base64 }" width="800"/>'
        )
        st.markdown(image_html, unsafe_allow_html=True)
    elif flag == "error":
        if "\n" in msg:
            with st.popover("Error Details"):
                st.error(msg.replace("\n", "  \n"))
        else:
            st.error(msg.replace("\n", "  \n"))
    elif flag == "success":
        st.success(msg.replace("\n", "  \n"))
    else:
        return False
    return True


def _display_chart(chart_id):
    import plotly.io as pio

    fig = pio.from_json(load_object(chart_id))
    st.plotly_chart(fig, use_container_width=True, key=chart_id)


def _display_flowchart(diagram_json):
    import json

    from streamlit_flow import streamlit_flow
    from streamlit_flow.elements import StreamlitFlowEdge, StreamlitFlowNode
    from streamlit_flow.layouts import LayeredLayout
    from streamlit_flow.state import StreamlitFlowState

    diagram = json.loads(diagram_json.strip())
    rand_id = diagram["rand_id"]
    prefix = "x" + rand_id + "_"
    chart_id = "chart_" + rand_id

    if chart_id not in st.session_state:
        nodes = []

        input_nodes = set(edge[0] for edge in diagram["edges"])
        output_nodes = set(edge[1] for edge in diagram["edges"])
        for node, tp in zip(diagram["nodes"], diagram["types"]):
            if node in input_nodes and node not in output_nodes:
                node_type = "input"
            elif node in output_nodes and node not in input_nodes:
                node_type = "output"
            else:
                node_type = "default"
            nodes.append(
                StreamlitFlowNode(
                    id=prefix + node,
                    pos=(0, 0),
                    data={"content": f"**{node}**\n\n({tp})"},
                    node_type=node_type,
                    source_position="right",
                    target_position="left",
                    draggable=True,
                )
            )

        edges = []
        for edge in diagram["edges"]:
            edge = StreamlitFlowEdge(
                id=prefix + edge[0] + "-" + edge[1],
                source=prefix + edge[0],
                target=prefix + edge[1],
                animated=False,
                marker_end={"type": "arrow"},
            )
            edges.append(edge)

        state = StreamlitFlowState(nodes, edges)
        st.session_state[chart_id] = state
    else:
        state = st.session_state[chart_id]

    streamlit_flow(
        chart_id + "_render",
        state,
        fit_view=True,
        layout=LayeredLayout(direction="right"),
        height=300,
        show_minimap=False,
        show_controls=True,
        pan_on_drag=True,
        allow_zoom=False,
        hide_watermark=True,
    )


def _display(markdown: str, visualize: bool):
    def _show_block(type: str, code: str):
        if type == "diagram":
            if visualize:
                try:
                    _display_flowchart(code)
                except Exception as e:
                    st.code(f"Error displaying diagram: {e}")
            else:
                st.code(code, language="json")
        elif type == "chart":
            if visualize:
                try:
                    _display_chart(code.strip())
                except Exception as e:
                    st.code(f"Error displaying chart: {e}")
            else:
                st.code(code)
        elif type == "editor":
            if visualize:
                try:
                    display_fv_editor(code.strip(), st.session_state.simulate_user_prompt_fn)
                except Exception as e:
                    st.code(f"Error displaying editor: {e}")
            else:
                st.code(code)
        else:
            st.code(code)

    display(markdown, st.markdown, _show_block)


class WorkLogHandler(UILogHandler):
    def __init__(self):
        super().__init__(diagram=False)
        self.records = []

    def emit_text(self, record) -> None:
        if _print_record(record):
            self.records.append(record)


if "copilot" not in st.session_state:
    args = parse_args()
    with Configs(
        llm=args["llm"], agent_invoke_kwargs={"max_iterations": 30}
    ).update_default():
        st.session_state.copilot = build_agent()
        print("Copilot initialized")

st.image(MELVIN_PATH, width=200)
st.title("Hi, I'm Melvin! Your Tecton Co-Pilot")

# Initialize chat history
if "messages" not in st.session_state:
    st.session_state.messages = []

chat_col, code_col = st.tabs(["Chat", "Source Code"])

with code_col:
    code_container = st.empty()
    set_code_container(code_container)

_code()

with chat_col:
    container = st.container()
    with container:
        # Display chat messages from history on app rerun
        last = len(st.session_state.messages) - 1
        last_time = None
        for i, message in enumerate(st.session_state.messages):
            avatar = MELVIN_PATH if message["role"] == "assistant" else None
            time = message["time"]
            with st.chat_message(message["role"], avatar=avatar):
                _display(message["content"], visualize=True)
                if "log" in message:
                    is_last = i == last
                    with st.expander("Actions", expanded=False):
                        for record in message["log"]:
                            _print_record(record)
                if message["role"] == "assistant":
                    response_sec = (message["time"] - last_time).total_seconds()
                    cost = message.get("cost", 0)
                    # with time, response time and cost
                    st.caption(
                        f"Time: {time:%Y-%m-%d %H:%M:%S},  Latency: {response_sec:.2f}s, Cost: ${cost:.4f}"
                    )
                else:
                    st.caption(f"Time: {time:%Y-%m-%d %H:%M:%S}")
            last_time = message["time"]


def on_submit(prompt_override: str = None):
    prompt = prompt_override or st.session_state.user_input
    chat_history = [(x["role"], x["content"]) for x in st.session_state.messages]
    st.session_state.messages.append(
        {"role": "user", "content": prompt, "time": datetime.now()}
    )

    # Display assistant response in chat message container
    with container:
        # Display user message in chat message container
        with st.chat_message("user"):
            _display(prompt, visualize=False)
        ct = st.chat_message("assistant", avatar=MELVIN_PATH)
        with ct:
            empty = st.empty()
            with empty.container():
                with st.status("Working on it ...", expanded=True, state="running"):
                    handler = WorkLogHandler()
                    with set_ui_logger(handler) as logger:
                        with st.session_state.copilot.set_logger(logger):
                            with cost_monitor() as monitor:
                                try:
                                    response = st.session_state.copilot.invoke(
                                        prompt, chat_history=chat_history
                                    )
                                except Exception as e:
                                    response = "Error encountered: " + str(e)
                    st.session_state.messages.append(
                        {
                            "role": "assistant",
                            "content": response,
                            "log": handler.records,
                            "cost": monitor.total_cost,
                            "time": datetime.now(),
                        }
                    )
            with empty.container():
                _display(response, visualize=False)

st.session_state.simulate_user_prompt_fn = on_submit

# React to user input
st.chat_input(
    "Ask anything about Tecton and your Tecton code ...",
    key="user_input",
    on_submit=on_submit,
)



import inspect
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from pydantic import BaseModel, Field, create_model
from tecton import (
    Attribute,
    DataSource,
    Entity,
    FeatureView,
    RequestSource,
    realtime_feature_view,
)
from tecton.framework.data_source import FilteredSource
from tecton.types import Field as TectonField
from tecton.types import String

from ..constants import _OUTPUT_JSON_FIELD, _TECTON_MOCK_OBJECT_ATTR
from ..utils._extra_attrs import get_attr, has_attr, set_attr
from ..utils.config_utils import Configs
from ._tecton_utils import make_request_source, set_conf
from .convert import tecton_type_to_python_annotation

_REQUEST = make_request_source(
    name=str, input=str, session_path=str, _dummy_request_ts=datetime
)


def is_local_source(source: Any) -> bool:
    src = _get_orig_source(source)
    return has_attr(src, _TECTON_MOCK_OBJECT_ATTR)


def get_local_source_attrs(sourcre: Any) -> Dict[str, Any]:
    src = _get_orig_source(sourcre)
    return get_attr(src, _TECTON_MOCK_OBJECT_ATTR)


def set_local_source_attrs(source: Any, attrs: Dict[str, Any]) -> None:
    src = _get_orig_source(source)
    set_attr(src, _TECTON_MOCK_OBJECT_ATTR, attrs)


def _get_orig_source(source: Any) -> DataSource:
    if isinstance(source, FilteredSource):
        return source.source
    if isinstance(source, DataSource):
        return source
    raise ValueError(f"{source} is not a DataSource or FilteredSource")


class FuncWrapper:
    def __init__(
        self,
        name: str,
        func: Callable,
        sources: Optional[List[FeatureView]] = None,
        input_schema: Optional[BaseModel] = None,
        assert_entity_defined: bool = False,
    ):
        self.name = name
        self.func = func
        self.sources = sources or []
        ap = _ArgumentParser(input_schema or func, self.sources, assert_entity_defined)
        (
            self.views,
            self.llm_args,
            self.feature_args,
            self.entity_args,
            self.has_request_source,
            self.should_add_request_source,
        ) = (
            ap.views,
            ap.llm_args,
            ap.feature_args,
            ap.entitye_args,
            ap.has_request_source,
            ap.should_add_request_source,
        )

    def make_prompt(self, **rtfv_kwargs: Any) -> FeatureView:
        return self.make_feature_view(**rtfv_kwargs)

    def make_feature_view(self, **rtfv_kwargs: Any) -> FeatureView:
        name = self.name
        func = self.func
        func_name = func.__name__
        doc = func.__doc__
        has_request_source = self.has_request_source
        rtfv_kwargs = Configs.get_default().get_rtfv_config(rtfv_kwargs)
        fv_names = [fv.name for fv in self.views]
        deco = realtime_feature_view(
            name=name,
            description=doc,
            sources=[_REQUEST] + self.views,
            mode="python",
            features=[Attribute(_OUTPUT_JSON_FIELD, String)],
            context_parameter_name="context",
            **rtfv_kwargs,
        )

        def service(request_context, *fvs, context=None):
            try:
                import json

                from ..constants import _OUTPUT_JSON_FIELD
                from ..utils.runtime import runtime_context

                if name != request_context["name"]:
                    return {_OUTPUT_JSON_FIELD: "{}"}
                if has_request_source:
                    pos_args: List[Any] = [json.loads(request_context["input"])]
                    params = {}
                else:
                    pos_args: List[Any] = []
                    params = json.loads(request_context["input"])
                try:  # make tecton compiler embed this function
                    f = func  # running locally
                except Exception:  # tecton runtime can't really get this function
                    f = globals()[func_name]  # running on server
                ctx = {"tecton_request": params.copy()}
                for key, fv in zip(fv_names, fvs):
                    params[key] = fv
                with runtime_context(
                    ctx,
                    tecton_context=context,
                    session_path=request_context.get("session_path"),
                ):
                    res = f(*pos_args, **params)
                    return {_OUTPUT_JSON_FIELD: json.dumps({"result": res})}
            except Exception:
                from traceback import format_exc

                return {_OUTPUT_JSON_FIELD: json.dumps({"error": format_exc()})}

        return self._wrap(deco, service)

    def _wrap(self, deco: Any, service: Any) -> Any:
        with set_serialization():
            if len(self.views) == 0:

                def _wrapper(request_context, context=None):
                    return service(request_context, context=context)

                return deco(_wrapper)
            if len(self.views) == 1:

                def _wrapper(request_context, fv1, context=None):
                    return service(request_context, fv1, context=context)

                return deco(_wrapper)
            if len(self.views) == 2:

                def _wrapper(request_context, fv1, fv2, context=None):
                    return service(request_context, fv1, fv2, context=context)

                return deco(_wrapper)
            if len(self.views) == 3:

                def _wrapper(request_context, fv1, fv2, fv3, context=None):
                    return service(request_context, fv1, fv2, fv3, context=context)

                return deco(_wrapper)
            if len(self.views) == 4:

                def _wrapper(request_context, fv1, fv2, fv3, fv4, context=None):
                    return service(request_context, fv1, fv2, fv3, fv4, context=context)

                return deco(_wrapper)
            if len(self.views) == 5:

                def _wrapper(request_context, fv1, fv2, fv3, fv4, fv5, context=None):
                    return service(
                        request_context, fv1, fv2, fv3, fv4, fv5, context=context
                    )

                return deco(_wrapper)
        raise NotImplementedError("Too many sources")


@contextmanager
def set_serialization():
    import os
    import sys

    if "ipykernel" in sys.modules or os.environ.get("TECTON_GEN_AI_DEV_MODE") == "true":
        yield
    else:  # not in notebook
        with set_conf({"TECTON_FORCE_FUNCTION_SERIALIZATION": "true"}):
            yield


def entity_to_pydantic(entity: Entity, fields: List[TectonField]) -> BaseModel:
    schema: Dict[str, str] = {}
    _fields = {item.name: item for item in fields}
    for name in entity.join_keys:
        schema[name] = (tecton_type_to_python_annotation(_fields[name].dtype), Field())
    return create_model(entity.name, **schema, __doc__=entity.description)


class _ArgumentParser:
    def __init__(
        self,
        func: Union[Callable, Type[BaseModel]],
        sources: List[Union[RequestSource, FeatureView]],
        assert_entity_defined: bool,
    ) -> Tuple[List[str], List[str], List[str], bool]:
        self.func = func
        if inspect.isclass(func) and issubclass(func, BaseModel):
            self.orig_args = list(func.model_fields.keys())
        else:
            self.orig_args = inspect.getfullargspec(func).args
        self.has_request_source = len(sources) > 0 and isinstance(
            sources[0], RequestSource
        )
        self.views: List[FeatureView] = (
            sources if not self.has_request_source else sources[1:]
        )
        if not self.has_request_source:
            self.llm_args, self.feature_args, self.entitye_args = (
                self._parse_arguments_no_request_sourcre(
                    self.orig_args,
                    assert_entity_defined,
                    assert_all_feature_args=False,
                )
            )
            self.view = sources
        else:
            _, self.feature_args, self.entitye_args = (
                self._parse_arguments_no_request_sourcre(
                    self.orig_args[1:],
                    assert_entity_defined,
                    assert_all_feature_args=True,
                )
            )
            self.llm_args = [x.name for x in sources[0].schema]
        self.should_add_request_source = (
            not self.has_request_source and len(self.llm_args) > 0
        )

    def _parse_arguments_no_request_sourcre(
        self,
        args: List[str],
        assert_entity_defined: bool,
        assert_all_feature_args: bool,
    ) -> Tuple[List[str], List[str], List[str]]:
        func = self.func
        feature_args = [source.name for source in self.views]
        if not set(feature_args).issubset(args):
            raise ValueError(
                f"All feature view names ({feature_args}) "
                f"must be defined in {func} arguments"
            )
        if len(self.views) == 0:
            entity_args: List[str] = []
        else:
            jk = self.views[0].join_keys
            for source in self.views[1:]:
                jk += source.join_keys
                # if set(jk) != set(source.join_keys):
                #    raise ValueError("All sources must have the same join keys")
            entity_args = list(dict.fromkeys(jk))
        if assert_entity_defined:
            if not set(entity_args).issubset(args):
                raise ValueError(
                    f"All entity keys ({entity_args}) must be defined in {func}"
                )
        llm_args = [x for x in args if x not in feature_args]
        if assert_all_feature_args and len(llm_args) > 0:
            raise ValueError(
                f"Only features and request source arguments are allowed: {llm_args}"
            )
        return llm_args, feature_args, entity_args


def assert_param_not_null_or_get_from_mock(obj, source, attr_name):
    if obj is not None:
        return obj
    if is_local_source(source):
        attrs = get_local_source_attrs(source)
        if attr_name in attrs:
            return attrs[attr_name]
    raise ValueError("Value can't be None")



from typing import Any, List

from constants import STARTING_DIR
from tecton import conf
from tecton._internals import metadata_service
from tecton_core.fco_container import FcoContainer
from tecton_proto.metadataservice.metadata_service__client_pb2 import (
    GetAllEntitiesRequest,
    GetAllFeatureServicesRequest,
    GetAllTransformationsRequest,
    GetAllVirtualDataSourcesRequest,
    GetTransformationRequest,
    ListWorkspacesRequest,
    QueryFeatureViewsRequest,
)
from utils import (
    _err,
    _log,
    _success,
    get_cwd,
    is_in_feature_repo,
    run_command,
    validate_syntax,
)


def create_feature_repo(repo_name: str) -> str:
    """
    Create a Tecton feature repo

    Args:

        repo_name: The name of the feature repo

    Returns:

        str: The output of the command
    """
    import os

    try:
        assert not is_in_feature_repo(), "You are already in a feature repo"

        cwd = get_cwd()
        repo_path = os.path.join(cwd, repo_name)
        os.makedirs(repo_path, exist_ok=False)
        os.chdir(repo_path)
        code, out, err = run_command("tecton init")
        if code == 0:
            return _success(
                f"{repo_name} created successfully, you are now in the repo: {repo_path}"
            )
        else:
            return _err(err)
    except Exception as e:
        return _err(e)


def move_to_folder(folder: str) -> str:
    """
    Move to a folder, a folder may or may not be a feature repo

    Args:

        folder: relative path or absolute path of the folder, if folder is empty, it will move back to the directory.

    Returns:

        str: whether the move is successful and whether it is a feature repo
    """
    import os

    try:
        if folder == "":
            os.chdir(STARTING_DIR)
            return _success("You are now in the starting directory " + STARTING_DIR)
        repo_path = os.path.abspath(folder)
        if not os.path.exists(repo_path):
            return _err(f"{repo_path} does not exist")
        if not os.path.isdir(repo_path):
            return _err(f"{repo_path} is not a directory")
        os.chdir(repo_path)
        is_fr = (
            "it is a feature repo"
            if is_in_feature_repo()
            else "it is not a feature repo"
        )
        return _success(f"You are now in {repo_path}, {is_fr}")
    except Exception as e:
        return _err(e)


def get_source_code_lineno() -> list[list[Any]]:
    """
    Get the python code (with line numbers starting from 1) for all data source and feature definitions

    Returns:

        list[list[str]]: The python code for all feature definitions,
            each element is also a list. It contains the line number as the first element
            and the code as the second element.
    """
    import os

    assert is_in_feature_repo(), "You must be in a feature repo"

    path = os.path.join(get_cwd(), "features.py")
    if not os.path.exists(path):
        res = []
    else:
        with open(path, "r") as f:
            lines = f.read().splitlines()
            res = [[i + 1, line] for i, line in enumerate(lines)]
    if len(res) == 0:
        res = [[1, ""]]
    if res[-1][1] != "":
        res.append([len(res) + 1, ""])
    return res


def get_source_code() -> str:
    """
    Get the python source code of features.py

    Returns:

        str: The python code of features.py
    """
    import os

    assert is_in_feature_repo(), "You must be in a feature repo"

    path = os.path.join(get_cwd(), "features.py")
    _log(":eyeglasses: Getting the source code from " + path)
    if not os.path.exists(path):
        return ""
    else:
        with open(path, "r") as f:
            return f.read()


def modify_source_code(
    modifications: dict[int, str] = {},
    insertions: dict[int, list[str]] = {},
    deletions: dict[int, int] = {},
    explanation: str = "",
) -> str:
    """
    Modify the code of data source and feature definitions, before each call to this tool,
    get_source_code must be used to get the latest original code,
    so the delta change can be calculated based on that.

    Args:

        modifications: A dictionary with line number as key and the new code as value, provide empty dict if no modification
        insertions: A dictionary with insertion line number as key and a list of new lines as value, provide empty dict if no insertion
        deletions: A dictionary with start line number as key and count of lines to delete as value, provide empty dict if no deletion
        explanation: An explanation of the changes, in under 100 words

    Returns:

        str: a message indicating whether the change is successful

    Note:

        To modify the code, it is always helpful to search tecton documents.
    """
    return _modify_source_code(
        modifications=modifications,
        insertions=insertions,
        deletions=deletions,
        explanation=explanation,
        revert_on_error=False,
    )


def fix_syntax_issues(
    modifications: dict[int, str] = {},
    insertions: dict[int, list[str]] = {},
    deletions: dict[int, int] = {},
    explanation: str = "",
) -> str:
    """
    Modify the code to fix syntax issues.
    get_source_code must be used to get the latest original code,
    so the delta change can be calculated based on that.

    Args:

        modifications: A dictionary with line number as key and the new code as value, provide empty dict if no modification
        insertions: A dictionary with insertion line number as key and a list of new lines as value, provide empty dict if no insertion
        deletions: A dictionary with start line number as key and count of lines to delete as value, provide empty dict if no deletion
        explanation: An explanation of the changes, in under 100 words

    Returns:

        str: a message indicating whether the change is successful

    Note:

        If the modification caused new syntax error, the change will be reverted and you should retry.
    """
    import os

    path = os.path.join(get_cwd(), "features.py")
    try:
        validate_syntax(path)
        return _success("No syntax issues found")
    except SyntaxError:
        return _modify_source_code(
            modifications=modifications,
            insertions=insertions,
            deletions=deletions,
            explanation="Syntax fixer: " + explanation,
            revert_on_error=True,
        )
    except Exception as e:
        return _err(e)


def save_modified_code(
    python_code: str, explanation: str = "", previous_problem: str = ""
) -> str:
    """
    Save the new feature definition source code.
    The new code must be the whole content of the file.

    Args:

        python_code: The new python code for the features.py
        explanation: An explanation of the changes, in under 100 words
        previous_problem: A summary of the previous problem leading to this change. Set this to
            non-empty string only when the change is to fix an issure reported by a validation.
            The summary should be under 100 words. If non-empty, it should start with "Previous problem: ". It should end with "(Gotcha identified)" if the problem was a gotcha.. It should end with "(tecton plan failed)" if tecton plan failed.

    Returns:

        str: The success or error message
    """
    import os

    if previous_problem:
        _log(":exclamation: " + previous_problem)
    _log(":writing_hand: " + explanation)

    path = os.path.join(get_cwd(), "features.py")
    if not python_code.endswith("\n"):
        python_code += "\n"
    with open(path, "w") as f:
        f.write(python_code)
    try:
        validate_syntax(path)
        run_command(f"black {path}")
    except SyntaxError as se:
        return _err(se, prefix="Syntax error")
    return _success("Feature definitions updated without syntax issues")


def validate_code_change() -> str:
    """
    Validate the current feature code.

    Returns:

        str: The validation result
    """
    try:
        import os

        assert is_in_feature_repo(), "You must be in in a feature repo"
        _log(":question: Validating the code change")
        path = os.path.join(get_cwd(), "features.py")
        validate_syntax(path)
        code, out, err = run_command("tecton plan")
        if code == 0:
            return _success(
                "The code has been successfully validated via `tecton plan`. No issues found."
            )
        else:
            return _err(f"{err}\n\n{out}")
    except Exception as e:
        return _err(e)


def _modify_source_code(
    modifications: dict[int, str] = {},
    insertions: dict[int, list[str]] = {},
    deletions: dict[int, int] = {},
    explanation: str = "",
    revert_on_error: bool = False,
) -> str:
    src = None
    try:
        import os

        src = get_source_code_lineno()

        if explanation:
            _log(explanation)

        dels = set()
        for start, count in deletions.items():
            start = int(start)
            dels.update(range(start, start + count))
        modifications = {int(k): v for k, v in modifications.items()}
        insertions = {int(k): v for k, v in insertions.items()}
        res = []
        if 0 in insertions:
            res += insertions[0]
        for line, code in src:
            if line in modifications:
                res.append(modifications[line])
            elif line in insertions:
                res += insertions[line]
                if line not in dels:
                    res.append(code)
            else:
                if line not in dels:
                    res.append(code)
        while line + 1 in insertions:
            res += insertions[line + 1]
            line += 1
        path = os.path.join(get_cwd(), "features.py")
        with open(path, "w") as f:
            f.write("\n".join(res))
        validate_syntax(path)
        run_command(f"black {path}")
        return _success("Feature definitions updated without syntax issues")
    except SyntaxError as se:
        if revert_on_error:
            with open(path, "w") as f:
                f.write("\n".join([x[1] for x in src]))
            return _err(
                f"{se}, the change has been reverted, try again!", prefix="Syntax error"
            )
        else:
            return _err(se, prefix="Syntax error")
    except Exception as e:
        return _err(e)


def tecton_cli_help() -> str:
    """
    Query the tecton cli for all of the commands that you can run with it
    Can be used to figure out which user you're logged in as, which tecton workspaces you have access to, which cluster you're connected to (by calling get-caller-identity) etc.

    Returns:

        str: list of commands you can run with the tecton cli
    """
    try:
        _log("Query tecton cli for available commands")
        code, out, err = run_command("tecton")
        if code != 0:
            return _err(f"{err}\n\n{out}")
        return out
    except Exception as e:
        return _err(e)


def tecton_cli_execute(command: str = "") -> str:
    """
    Execute a tecton cli command
    Use the tecton_cli_help tool to figure out which commands you have at your disposal

    Args:

        command: tecton command to execute, including any flags for that command. Do not prefix with the name of the cli, tecton

    Returns:

        str: The result of the command. May indicate success or failure
    """
    try:
        _log(f"Running tecton cli command {command}")

        code, out, err = run_command(f"tecton {command}")
        if code != 0:
            return _err(f"{err}\n\n{out}")
        return out
    except Exception as e:
        return _err(e)


def find_tecton_feature_repositories() -> List[str]:
    """
    Find available tecton feature repositories (sometimes also just called tecton repos or feature repos)

    Returns:

        List[str]: A list of paths to available feature repositories
    """
    import os

    tecton_folders = []

    for root, dirs, files in os.walk(os.getcwd()):
        if ".tecton" in files:
            tecton_folders.append(os.path.abspath(root))

    return tecton_folders


def list_workspaces() -> List[str]:
    """
    List all workspaces in the currently connected Tecton cluster.

    Returns:
        List[str]: A list of workspace names in the connected Tecton cluster.
        Returns an empty list if the operation fails.
    """
    try:
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)
        return [ws.name for ws in response.workspaces]
    except Exception as e:
        _err(f"Failed to list workspaces: {str(e)}")
        return []


def list_feature_views() -> List[dict]:
    """
    List all feature views in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each feature view
    """
    try:
        feature_views = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = QueryFeatureViewsRequest(workspace=workspace)
            response = metadata_service.instance().QueryFeatureViews(request)

            for fco in response.fco_container.fcos:
                if hasattr(fco, "feature_view") and fco.feature_view.fco_metadata.name:
                    feature_views.append(
                        {
                            "name": fco.feature_view.fco_metadata.name,
                            "description": fco.feature_view.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(fco.feature_view.fco_metadata.tags),
                        }
                    )

        return sorted(feature_views, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list feature views: {str(e)}")
        return []


def list_feature_services() -> List[dict]:
    """
    List all feature services in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each feature service
    """
    try:
        feature_services = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllFeatureServicesRequest(workspace=workspace)
            response = metadata_service.instance().GetAllFeatureServices(request)

            for feature_service in response.feature_services:
                if feature_service.fco_metadata.name:
                    feature_services.append(
                        {
                            "name": feature_service.fco_metadata.name,
                            "description": feature_service.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(feature_service.fco_metadata.tags),
                        }
                    )

        return sorted(feature_services, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list feature services: {str(e)}")
        return []


def list_transformations() -> List[dict]:
    """
    List all transformations in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each transformation
    """
    try:
        transformations = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllTransformationsRequest(workspace=workspace)
            response = metadata_service.instance().GetAllTransformations(request)

            for transformation in response.transformations:
                if transformation.fco_metadata.name:
                    transformations.append(
                        {
                            "name": transformation.fco_metadata.name,
                            "description": transformation.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(transformation.fco_metadata.tags),
                        }
                    )

        return sorted(transformations, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list transformations: {str(e)}")
        return []


def get_feature_view_code(name: str, workspace: str) -> str:
    """
    Get the code definition of a feature view from the Tecton cluster.
    This essentially gets the transformation using the feature view name and assumes its the same name as transformation name which works in most cases.

    Args:
        name: The name of the feature view to retrieve
        workspace: The workspace containing the feature view

    Returns:
        str: The Python code definition of the feature view

    Raises:
        Exception: If the feature view is not found or other errors occur
    """
    try:
        request = GetTransformationRequest(
            name=name,
            workspace=workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetTransformation(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        transformation_spec = fco_container.get_single_root()

        if transformation_spec is None:
            msg = f"Transformation '{name}' not found in workspace '{workspace}'. Try running `list_feature_views()` to view all registered feature views."
            raise ValueError(msg)

        # Validate that all required fields are present
        if not hasattr(transformation_spec, "validation_args"):
            raise ValueError("Missing validation_args in transformation_spec")
        if not hasattr(transformation_spec.validation_args, "transformation"):
            raise ValueError("Missing transformation in validation_args")
        if not hasattr(transformation_spec.validation_args.transformation, "args"):
            raise ValueError("Missing args in transformation")
        if not hasattr(
            transformation_spec.validation_args.transformation.args, "user_function"
        ):
            raise ValueError("Missing user_function in args")
        if not hasattr(
            transformation_spec.validation_args.transformation.args.user_function,
            "body",
        ):
            raise ValueError("Missing body in user_function")

        return (
            transformation_spec.validation_args.transformation.args.user_function.body
        )
    except Exception as e:
        _err(f"Failed to get feature view code: {str(e)}")
        return ""


def list_data_sources() -> List[dict]:
    """
    List all data sources in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each data source
    """
    try:
        data_sources = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllVirtualDataSourcesRequest(workspace=workspace)
            response = metadata_service.instance().GetAllVirtualDataSources(request)

            for data_source in response.virtual_data_sources:
                if data_source.fco_metadata.name:
                    data_sources.append(
                        {
                            "name": data_source.fco_metadata.name,
                            "description": data_source.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(data_source.fco_metadata.tags),
                        }
                    )

        return sorted(data_sources, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list data sources: {str(e)}")
        return []


def list_entities() -> List[dict]:
    """
    List all entities in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each entity
    """
    try:
        entities = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllEntitiesRequest(workspace=workspace)
            response = metadata_service.instance().GetAllEntities(request)

            for entity in response.entities:
                if entity.fco_metadata.name:
                    entities.append(
                        {
                            "name": entity.fco_metadata.name,
                            "description": entity.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(entity.fco_metadata.tags),
                        }
                    )

        return sorted(entities, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list entities: {str(e)}")
        return []



from ..constants import TECTON_DEV_MODE_KEY
import os


def set_dev_flag():
    os.environ[TECTON_DEV_MODE_KEY] = "true"


def is_dev_mode():
    return os.environ.get(TECTON_DEV_MODE_KEY, "false") == "true"



import importlib
import inspect
from functools import lru_cache
from typing import Any, Iterable, List, Tuple

from ...api import Agent, Configs, tool
from ...utils.code_parser import get_declaration

_DEFAULT_GEN_AI_MODULES = tuple(
    [
        "tecton_gen_ai.api",
        "tecton_gen_ai.testing",
    ]
)


@lru_cache
def make_copilot(
    modules: Tuple = _DEFAULT_GEN_AI_MODULES,
    mode: str = "brute_force",
    llm: Any = None,
    name: str = "copilot",
) -> Agent:
    prefix = """You are a copilot that helps users use tecton_gen_ai package to build AI projects.
Your main task is generating code based on user's input."""
    objects = list(_FuncOrClass.from_expressions(modules))
    if (llm or Configs.get_default().llm) is None:
        raise ValueError("Please set the language model for the copilot.")
    if mode == "brute_force":
        declarations = "\n\n".join([obj.declaration for obj in objects])

        sys_prompt = f"""{prefix}

Here are the functions and classes you can use to generate code:

{declarations}
"""

        return Agent(name=name, prompt=sys_prompt, llm=llm)
    elif mode == "tools":
        tools = [obj.to_tool() for obj in objects]
        return Agent(name=name, prompt=prefix, tools=tools, llm=llm)
    else:
        raise ValueError(f"Invalid mode: {mode}")


def assist(
    question: str,
    modules: Tuple = _DEFAULT_GEN_AI_MODULES,
    llm: Any = None,
    mode: str = "brute_force",
) -> Any:
    """
    Get the copilot's response to the user's question.

    TODO: not using tools seem to provide much better results.
    """
    from IPython.display import Markdown, display

    agent = make_copilot(modules=tuple(modules), mode=mode, name="copilot", llm=llm)
    resp = agent.invoke(question)
    display(Markdown(resp))


def chat(
    llm: Any = None,
    mode: str = "brute_force",
    modules: Tuple = _DEFAULT_GEN_AI_MODULES,
) -> None:
    """
    Chat with the copilot in notebook to generate code based on user's input.
    """

    from ...interactive import chat as _chat

    agent = make_copilot(modules=tuple(modules), mode=mode, name="copilot", llm=llm)
    _chat(agent)


class _FuncOrClass:
    def __init__(self, obj: Any):
        self.obj = obj
        self.declaration = get_declaration(obj, entrypoint_only=False)
        self.callable_declaration = get_declaration(obj, entrypoint_only=True)
        self.name = obj.__name__
        self.doc = (obj.__doc__ or "").split("Args:")[0].strip()

    @property
    def tool_description(self) -> str:
        return (
            f"Get the declaration of {self.name}. "
            f"The purpose of {self.name}:\n{self.doc}"
        )

    @property
    def tool_name(self) -> str:
        return f"get_declaration_of_{self.name}"

    def to_tool(self) -> Any:
        description = self.tool_description
        name = self.tool_name

        @tool(name=name, description=description)
        def _tool() -> str:
            return self.declaration

        return _tool

    @staticmethod
    def from_expressions(expressions: List[str]) -> Iterable["_FuncOrClass"]:
        for expr in expressions:
            try:
                module = importlib.import_module(expr)
                names = [name for name in dir(module) if not name.startswith("_")]
            except Exception:
                parts = expr.rsplit(".", 1)
                module = importlib.import_module(parts[0])
                names = [parts[1]]
            objs = [getattr(module, name) for name in names]
            yield from [
                _FuncOrClass(x)
                for x in objs
                if inspect.isclass(x) or inspect.isfunction(x)
            ]



import logging
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Dict, Iterator, List, Optional

import ipywidgets as widgets
from IPython.display import display

from ..utils.log import NOOP_LOGGER


class UILogHandler(logging.Handler):
    def __init__(self, diagram: bool):
        super().__init__()
        self.diagram = diagram
        self.history: List[List[str]] = []
        self.prompt: Dict[str, Any] = {}

    def emit(self, record: Any) -> None:
        if not self.diagram:
            self.emit_text(record)
        else:
            flow_event = getattr(record, "flow_event", None)
            if flow_event is not None:
                if flow_event.get("type") == "prompt":
                    self.prompt = flow_event
                    return
                diagram = self.build_diagram(flow_event)
                self.emit_diagram(diagram=diagram)

    def emit_text(self, record: Any) -> None:
        raise NotImplementedError

    def emit_diagram(self, diagram: Any) -> None:
        raise NotImplementedError

    def build_diagram(self, flow_event: Any) -> None:
        from .diagrams import plot_execution

        if flow_event["type"] == "llm":
            self.history.append([])
        else:
            self.history[-1].append(flow_event)
        prompt_uses_features = len(self.prompt.get("source_names", [])) > 0
        return plot_execution(self.history, prompt_uses_features)


class _WidgetLogHandler(UILogHandler):
    def __init__(self, output: widgets.Output, diagram: bool):
        super().__init__(diagram)
        self.output = output

    def emit_text(self, record: Any) -> None:
        with self.output:
            print(self.format(record))

    def emit_diagram(self, diagram: Any) -> None:
        with self.output:
            self.output.clear_output()
            display(diagram)


_UI_LOGGER = ContextVar("ui_logger", default=NOOP_LOGGER)


@contextmanager
def set_ui_logger(hanlder: Optional["UILogHandler"]) -> Iterator[logging.Logger]:
    if hanlder is None:
        logger = NOOP_LOGGER
    else:
        logger = logging.getLogger("widget")
        logger.handlers.clear()
        logger.addHandler(hanlder)
        logger.propagate = False
        logger.setLevel(logging.DEBUG)
    token = _UI_LOGGER.set(logger)
    try:
        yield logger
    finally:
        _UI_LOGGER.reset(token)


def get_ui_logger() -> logging.Logger:
    return _UI_LOGGER.get()



# flake8: noqa

from .agent.base import AgentBase, AgentInputModel, FeatureServiceConfig, VectorDBConfig
from .agent.client import get_agent
from .agent.service import Agent, fv_as_tool
from .utils.vector_db import VectorDB
from .tecton_utils.deco import emitter, prompt, tool
from .tecton_utils.extraction import llm_extraction
from .tecton_utils.knowledge import source_as_knowledge
from .utils.config_utils import Configs
from .utils.llm import LLMGenerationConfig
from .utils.runtime import RuntimeVar, get_runtime_var, get_session_var, set_session_var



import json
from datetime import datetime
from typing import Any, Dict, Sequence, Tuple, Union
from urllib.parse import urljoin

from requests.models import Response
from tecton_core import http


def ingest(
    cluster_url: str,
    workspace: str,
    api_key: str,
    push_source_name: str,
    ingestion_records: Sequence[Dict[str, Any]],
    timeout_sec: float,
) -> Dict[str, Any]:
    if not cluster_url.startswith("https://preview."):
        cluster_url = cluster_url.replace("https://", "https://preview.")
    status_code, reason, response = IngestionClient(
        cluster_url=cluster_url, api_key=api_key
    ).ingest(
        workspace_name=workspace,
        push_source_name=push_source_name,
        ingestion_records=ingestion_records,
        timeout_sec=timeout_sec,
    )
    if status_code >= 500:
        raise Exception(json.dumps(response))
    elif status_code >= 400:
        raise Exception(
            json.dumps(dict(status_code=status_code, reason=reason, response=response))
        )
    else:
        return response


# TODO: This is a temporary solution until TectonClient is available
# This is modified version from tecton._internals.ingestion.IngestionClient
# The reason for this duplication:
# 1. The original class is inside tecton library, and dependency on tecton library is not necessary and causes problems
# 2. This logic should run both locally or inside a RTFV
# 3. This logic should be inside TectonClient, but it is not available now
# https://linear.app/tecton/issue/FE-2663/add-steam-ingest-api-support-to-python-http-client
class IngestionClient:
    def __init__(self, cluster_url: str, api_key: str):
        self.ingestion_url = urljoin(cluster_url, "ingest")
        self.api_key = api_key

    def ingest(
        self,
        workspace_name: str,
        push_source_name: str,
        ingestion_records: Sequence[Dict[str, Any]],
        timeout_sec: float,
        dry_run: bool = False,
    ) -> Tuple[int, str, Union[Dict[str, Any], None]]:
        http_request = {
            "workspace_name": workspace_name,
            "dry_run": dry_run,
            "records": {
                push_source_name: [
                    {
                        "record": self._serialize_record(ingestion_record),
                    }
                    for ingestion_record in ingestion_records
                ]
            },
        }
        response = http.session().post(
            self.ingestion_url,
            json=http_request,
            headers=self._prepare_headers(),
            timeout=timeout_sec,
        )
        return self._extract_from_response(response)

    @staticmethod
    def _extract_from_response(
        response: Response,
    ) -> Tuple[int, str, Union[Dict[str, Any], None]]:
        # NOTE: Responses not coming from our system will not necessarily be in JSON format

        if response.status_code != 200:
            error_message = f"Failed to ingest records: {response.reason}"
            try:
                response_data = response.json()
            except ValueError:
                response_data = None
                error_message += f"\nAdditionally, response content is not valid JSON: {response.text}"
            return response.status_code, error_message, response_data

        try:
            response_data = response.json()
        except ValueError:
            error_message = f"Response content is not valid JSON: {response.text}"
            return response.status_code, error_message, None
        return response.status_code, response.reason, response_data

    def _prepare_headers(self) -> Dict[str, str]:
        return {
            "authorization": f"Tecton-key {self.api_key}",
            "Content-Type": "application/json",
        }

    @staticmethod
    def _serialize_record(record: Dict[str, Any]) -> Dict[str, Any]:
        serialized_record = {}
        for k, v in record.items():
            if isinstance(v, datetime):
                serialized_record[k] = v.isoformat()
            else:
                serialized_record[k] = v
        return serialized_record



import pandas as pd
from typing import List, Dict, Any
from pandas.api.types import is_datetime64_any_dtype


def to_records(
    df: pd.DataFrame, timestamp_to: str = "original"
) -> List[Dict[str, Any]]:
    """
    Convert a DataFrame to a list of dictionaries

    Args:

        df: The DataFrame
        timestamp_to: The timestamp field to convert to, defaults to "original",
            which will keep the original pandas Timestamp. When set to "py", it will
            convert the timestamp to a Python datetime object. When set to "str", it
            will convert the timestamp to a string.

    Returns:

        List[Dict[str,Any]]: The list of dictionaries
    """
    cols: Dict[str, pd.Series] = {}
    for col in df.columns:
        s = df[col]
        if is_datetime64_any_dtype(s):
            if timestamp_to == "original":
                cols[col] = s
            elif timestamp_to == "py":
                cols[col] = pd.Series(s.dt.to_pydatetime(), dtype=object)
            elif timestamp_to == "str":
                cols[col] = df[col].dt.strftime("%Y-%m-%d %H:%M:%S")
            else:
                raise ValueError(f"Unsupported timestamp_to {timestamp_to}")
        else:
            cols[col] = df[col]
    return pd.DataFrame(cols).to_dict(orient="records")



from pydantic import Field
from utils import _err, _log, _log_with_details, load_object, store_object

from melvin.tecton_gen_ai.api import Agent, AgentInputModel


class Input(AgentInputModel):
    query: str = Field(
        description="The detailed instruction on how to generate the chart"
    )
    input_session_object_key: str = Field(
        description="The session object key of input data from calling a dataframe generator, the key must start with `so_df_`",
    )


def sys_prompt(input_session_object_key: str) -> str:
    import datetime

    try:
        description = load_object(input_session_object_key)["description"]
    except Exception as e:
        return (
            f"There is an error with the input data key: {input_session_object_key}. Error: {e} "
            "Do not call any tool, directly return this error message"
        )
    time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return f"""
You are going to generate a single chart based on user's instruction and the input data.

You need to first transform the data using duckdb SQL by `run_sql`.

The data transformation should take the chart type into consideration

After the transformation, call one and only one chart tool based on the user's intent to output the chart image.

If not successful, you should return an error message starting with `Error: `
If successful, return a chart representation which is a session object key starting with `so_chart_`,
and surround it with triple backticks, language is `tecton_chart`

The final chart must be generated by one of the chart tools. don't manipulate the `so_chart_` id.

Example of the chart representation:

```tecton_chart
so_chart_sdgg
```

Note:

    - strptime is for string to time conversion
    - strftime is for time to string conversion
    - a period of time from now (e.g. since 3 days ago) can be expressed as: `current_localtimestamp() - INTERVAL 3 DAY`

When the chart type is not explicitly specified in the query, you should choose the chart type based on these guidelines:

    - Line chart: for timeseries, and each line represents a column of the data
    - Line chart with categories: for timeseries, and each line represents a value in a category column of the data
    - Stacked area chart: for timeseries, similar as line chart with categories but when a breakdown is specified
    - Pie chart: not for timeseries, but a breakdown on values of a column

Current time:

{time}

The generated duckdb SQL should only use the columns in the description and schema of the input data.
Here are the description and schema of the input data:

{description}
"""


def run_sql(input_session_object_key: str, duckdb_sql: str) -> str:
    """
    Run duckdb SQL to transform the data before calling a chart tool.
    It should assume the query is operating on a table with name: `input_data`
    And the description and schema of `input_data` is the same as the input data

    Args:

        input_session_object_key: the key to the input data
        duckdb_sql: the duckdb_sql to run, the sql must be a SELECT statement

    Returns:

        str: the session object key of the transformed data, starting with `so_charttemp_` or an error message
    """
    import duckdb

    _log_with_details(
        ":question: Transforming data using duckdb", f"```sql\n{duckdb_sql}"
    )
    try:
        input_data = load_object(input_session_object_key)["df"]  # noqa
        input_row = len(input_data)
        with duckdb.connect() as con:
            output_df = con.sql(duckdb_sql).df()
            output_row = len(output_df)
        _log(
            "Data transformed. "
            f"Input data has {input_row} rows, output data has {output_row} rows"
        )
        return store_object(output_df, "charttemp")
    except Exception as e:
        return _err(e)


def generate_line_chart(
    transformed_data_key: str, x_label: str, y_labels: list[str], title: str
) -> str:
    """
    Generate line chart. `run_sql` must be called to transform the original data

    Args:

        transformed_data_key: the session object key generated by `run_sql`, starting with `so_charttemp_`
        x_label: label of x axis
        y_labels: list of labels of the lines
        title: the title of the chart

    Returns:

        str: the session object key of the output image, starting with `so_chart_`

    Note:

        - The first column should be the x axis
        - The rest columns will be on the y axis
    """
    import plotly.express as px
    import plotly.io as pio

    _log(f":chart: Generating line chart: {title}")

    df = load_object(transformed_data_key)
    df.columns = [x_label] + y_labels
    fig = px.line(
        df,
        x=df.columns[0],
        y=df.columns[1:],
        title=title,
    )
    return store_object(pio.to_json(fig), "chart")


def generate_line_chart_with_categories(
    transformed_data_key: str,
    x_label: str,
    y_label: str,
    category_label: str,
    title: str,
) -> str:
    """
    Generate line chart with the category where the categories are also from the dataframe.
    `run_sql` must be called to transform the original data

    Args:

        transformed_data_key: the session object key generated by `run_sql`, starting with `so_charttemp_`
        x_label: label of x axis
        y_label: label of y axis
        category_label: label of category axis
        title: the title of the chart

    Returns:

        str: the session object key of the output image, starting with `so_chart_`

    Note on the data transformation:

        - The first column should be the x axis
        - The last column should be the value given each x and categories
        - The middle columns should be the category columns

    """
    import pandas as pd
    import plotly.express as px
    import plotly.io as pio

    _log(f":chart: Generating line chart with categories: {title}")
    df = load_object(transformed_data_key)
    if len(df.columns) < 3:
        return _err("There should be at least 3 columns for this chart")
    elif len(df.columns) == 3:
        df.columns = [x_label, category_label, y_label]
    else:
        str_col = df[df.columns[1]].astype(str)
        for i in range(2, len(df.columns) - 1):
            str_col = str_col.str.cat(df[df.columns[i]].astype(str), sep=" - ")
        df = pd.DataFrame(
            {
                x_label: df[df.columns[0]],
                category_label: str_col,
                y_label: df[df.columns[-1]],
            }
        )
    fig = px.line(
        df,
        x=df.columns[0],
        y=df.columns[-1],
        color=df.columns[1],
        title=title,
    )
    return store_object(pio.to_json(fig), "chart")


def generate_stacked_area_chart(
    transformed_data_key: str,
    x_label: str,
    y_label: str,
    category_label: str,
    title: str,
) -> str:
    """
    Generate stacked area chart with the category where the categories are also from the dataframe.
    `run_sql` must be called to transform the original data

    Args:

        transformed_data_key: the session object key generated by `run_sql`, starting with `so_charttemp_`
        x_label: label of x axis
        y_label: label of y axis
        category_label: label of category axis
        title: the title of the chart

    Returns:

        str: the session object key of the output image, starting with `so_chart_`

    Note on the data transformation:

        - The first column should be the x axis
        - The last column should be the value given each x and categories
        - The middle columns should be the category columns

    """
    import pandas as pd
    import plotly.express as px
    import plotly.io as pio

    _log(f":chart: Generating stacked area chart: {title}")
    df = load_object(transformed_data_key)
    if len(df.columns) < 3:
        return _err("There should be at least 3 columns for this chart")
    elif len(df.columns) == 3:
        df.columns = [x_label, category_label, y_label]
    else:
        str_col = df[df.columns[1]].astype(str)
        for i in range(2, len(df.columns) - 1):
            str_col = str_col.str.cat(df[df.columns[i]].astype(str), sep=" - ")
        df = pd.DataFrame(
            {
                x_label: df[df.columns[0]],
                category_label: str_col,
                y_label: df[df.columns[-1]],
            }
        )
    fig = px.area(
        df,
        x=df.columns[0],
        y=df.columns[-1],
        color=df.columns[1],
        title=title,
    )
    return store_object(pio.to_json(fig), "chart")


def generate_pie_chart(transformed_data_key: str, title: str) -> str:
    """
    Generate pie chart. `run_sql` must be called to transform the original data

    Args:

        transformed_data_key: the session object key generated by `run_sql`, starting with `so_charttemp_`
        title: the title of the chart

    Returns:

        str: the session object key of the output image, starting with `so_chart_`

    Note:

        - The first column should be the labels
        - The second column should be the values
    """
    import plotly.express as px
    import plotly.io as pio

    _log(f":chart: Generating pie chart: {title}")

    df = load_object(transformed_data_key)
    if len(df.columns) != 2:
        return _err("There should be exactly 2 columns for this chart")
    fig = px.pie(df, names=df.columns[0], values=df.columns[1], title=title)
    return store_object(pio.to_json(fig), "chart")


def _make_chart_agent() -> Agent:
    return Agent(
        name="chart_agent",
        description="""An agent that generates charts given instruction and input data

Returns:

    if not successful, it returns an error message starting with `Error: `
    if successful, it returns the chart type (e.g. line chart) and a chart representation quoted with triple backticks, language is `tecton_chart`

Note:

    - A dataframe generator tool must be called before calling this agent
    - The output of this agent should never be used to call this agent again
    - For any follow up question, the input data needs to be regenerated
    - Do not try to interpret the content in a tecton_chart code block, use the exact code block in the final output.
""",
        prompt=sys_prompt,
        input_schema=Input,
        tools=[
            run_sql,
            generate_line_chart,
            generate_line_chart_with_categories,
            generate_pie_chart,
            generate_stacked_area_chart,
        ],
    )



# flake8: noqa
from .mocks import (
    make_local_batch_feature_view,
    make_local_realtime_feature_view,
    make_local_source,
    make_local_stream_feature_view,
)
from .utils import make_local_vector_db_config, print_md, set_dev_mode

try:
    from .examples.copilot import assist
except ImportError:
    pass



import pathlib
from typing import Any, List
from uuid import uuid4

from diagrams import Diagram, Edge
from diagrams.custom import Custom

_ICON_PATH = pathlib.Path(__file__).parent.parent.resolve() / "resources"
_DEFUALT_GRAPH_ATTR = {
    "pad": "0",
    "size": "10,6",
    "ratio": "0.3",
    "splines": "line",
}
_DEFUALT_NODE_ATTR = {
    "fontsize": "15",
    "fixedsize": "true",
}


class LLM(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "llm.png"), **kwargs)


class Input(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "input.png"), **kwargs)


class Output(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "input.png"), **kwargs)


class Prompt(Custom):
    def __init__(self, label: str, enriched: bool, **kwargs: Any):
        suffix = "-enriched" if enriched else ""
        super().__init__(label, str(_ICON_PATH / f"prompt{suffix}.png"), **kwargs)


class Tool(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "tool.png"), **kwargs)


class Table(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "table.png"), **kwargs)


class VectorDB(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "vdb.png"), **kwargs)


class FeatureTool(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "feature-tool.png"), **kwargs)


class KnowledgeTool(Custom):
    def __init__(self, label: str, **kwargs: Any):
        super().__init__(label, str(_ICON_PATH / "knowledge-tool.png"), **kwargs)


def plot_execution(
    history: List[List[Any]],
    prompt_uses_features: bool,
    graph_attr=_DEFUALT_GRAPH_ATTR,
    node_attr=_DEFUALT_NODE_ATTR,
) -> Diagram:
    def get_id():
        return str(uuid4())

    def edge(**kwargs: Any):
        return Edge(penwidth="2", **kwargs)

    class QuietDiagram(Diagram):
        def _repr_png_(self):
            return self.dot.pipe(format="png", quiet="true")

    with QuietDiagram(
        "", show=False, graph_attr=graph_attr, node_attr=node_attr, direction="LR"
    ) as diag:
        prompt = Prompt("System Prompt", prompt_uses_features)
        query = Input("Input")
        end = [query, prompt]
        for x in history:
            if len(x) == 0:
                new_end = LLM("LLM", id=get_id())
                end >> edge() >> new_end
            else:
                tools = []
                for tc in x:
                    tp = Tool
                    if tc["subtype"] == "fv":
                        tp = Table
                    elif tc["subtype"] == "retriever":
                        tp = VectorDB
                    elif tc.get("features", []):
                        tp = FeatureTool
                    elif "knowledge" in tc:
                        tp = KnowledgeTool
                    tool = tp(tc["value"], id=get_id())
                    tools.append(tool)
                new_end = tools
                if len(new_end) == 1:
                    new_end = new_end[0]
                end >> edge() >> LLM("LLM", id=get_id()) >> edge() >> new_end
            end = new_end
        end >> edge() >> Output("Output")

    return diag



#!/usr/bin/env python
import argparse
import sys
import streamlit.web.cli as stcli
import os

def main():
    parser = argparse.ArgumentParser(description="Melvin - Tecton Copilot")
    parser.add_argument("--port", type=int, default=3000, help="Port to run the server on")
    parser.add_argument("--host", type=str, default="localhost", help="Host to run the server on")
    parser.add_argument("--no-browser", action="store_true", help="Don't open browser automatically")
    
    args = parser.parse_args()
    
    # Get the absolute path to the UI file
    ui_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ui.py")
    
    # Prepare Streamlit arguments
    sys.argv = [
        "streamlit", "run", 
        ui_path,
        "--server.port", str(args.port),
        "--server.address", args.host,
        "--browser.gatherUsageStats", "false",
        "--theme.base", "light",
        "--global.showAppInBrowser", "You can now view Melvin in your browser"
    ]
    
    if args.no_browser:
        sys.argv.extend(["--server.headless", "true"])
    
    # Run Streamlit
    sys.exit(stcli.main())

if __name__ == "__main__":
    main()



import streamlit as st
from editors.feature_view_editor_ui import FeatureEditorState
from editors import feature_view_editor_ui  # Import the module with our function

# Your existing app setup
st.set_page_config(page_title="My Main App", layout="wide")
st.title("My Main Application")

# Your existing content...

st.session_state['editor_state'] = st.session_state.get('editor_state', FeatureEditorState())

# Embed the Tecton Feature Creator where needed
col1, col2 = st.columns(2)
with col1:
    generated_code = feature_view_editor_ui.display_editor(
        key_prefix="tecton_editor",
        state=st.session_state['editor_state']
    )

    if generated_code:
        # Do something with the generated code if needed
        st.session_state.last_generated_code = generated_code



import os
import tempfile
from typing import Any, Union
from unittest.mock import Mock, patch

import pytest
from pydantic import BaseModel

from melvin.tecton_gen_ai.utils._cache import CacheProtocol, get_cache
from melvin.tecton_gen_ai.utils.config_wrapper import from_json_config, to_json_config
from melvin.tecton_gen_ai.utils.runtime import RuntimeVar
from melvin.tecton_gen_ai.utils.structured_outputs import (
    BatchProcessor,
    T_OutputType,
    batch_generate,
    batch_generate_dicts,
)

MODELS = [
    "openai/gpt-4o-mini",
    {
        "model": "anthropic/claude-3-5-sonnet-20241022",
        "max_tokens": 100,
        "api_key": RuntimeVar(name="ANTHROPIC_API_KEY"),
    },
]


def generate_dict(
    model: Union[str, dict[str, Any]],
    text: str,
    schema: Union[type[T_OutputType], dict[str, Any]],
) -> dict[str, Any]:
    return batch_generate_dicts(model, [text], schema)[0]


def generate(
    model: Union[str, dict[str, Any]], text: str, schema: type[T_OutputType]
) -> T_OutputType:
    return batch_generate(model, [text], schema)[0]


class SampleModel(BaseModel):
    field1: str
    field2: int


@pytest.fixture(autouse=True)
def reset_cache():
    with tempfile.TemporaryDirectory() as temp_dir:
        _old_cache_dir = os.environ.get("XDG_CACHE_HOME")
        os.environ["XDG_CACHE_HOME"] = str(temp_dir)
        try:
            yield
        finally:
            if _old_cache_dir is not None:
                os.environ["XDG_CACHE_HOME"] = _old_cache_dir


@pytest.mark.parametrize("model", MODELS)
def test_generate(model):
    result = generate(model, "field1: value, field2: 42", SampleModel)
    assert result.field1 == "value"
    assert result.field2 == 42


@pytest.mark.parametrize("model", MODELS)
def test_generate_dict(model):
    result = generate_dict(model, "field1: value, field2: 42", SampleModel)
    expected = {"field1": "value", "field2": 42}
    assert result == expected


@pytest.mark.parametrize("model", MODELS)
def test_batch_generate(model):
    results = batch_generate(
        model,
        ["field1: value1, field2: 42", "field1: value2, field2: 43"],
        SampleModel,
    )
    expected = [
        SampleModel(field1="value1", field2=42),
        SampleModel(field1="value2", field2=43),
    ]
    assert len(results) == len(expected)
    for result, expected in zip(results, expected):
        assert result.field1 == expected.field1
        assert result.field2 == expected.field2


@pytest.mark.parametrize("model", MODELS)
def test_batch_generate_dicts(model):
    results = batch_generate_dicts(
        model,
        ["field1: value1, field2: 42", "field1: value2, field2: 43"],
        SampleModel,
        enable_cache=False,
    )
    expected = [{"field1": "value1", "field2": 42}, {"field1": "value2", "field2": 43}]
    assert len(results) == len(expected)
    for result, expected in zip(results, expected):
        assert result == expected


def test_batch_generate_dicts_invalid_model():
    with pytest.raises(ValueError, match="Model .* not supported"):
        batch_generate_dicts("invalid-model", ["text1", "text2"], SampleModel)


def test_batch_processor():
    processor = BatchProcessor.make("openai/gpt-4o-mini", SampleModel)
    processor: BatchProcessor = from_json_config(to_json_config(processor))
    assert processor.batch_process(
        ["field1: value1, field2: 42", "field1: value2, field2: 43"]
    ) == [
        dict(field1="value1", field2=42),
        dict(field1="value2", field2=43),
    ]
    assert [
        m.model_dump()
        for m in processor.batch_process(
            ["field1: value1, field2: 42", "field1: value2, field2: 43"],
            keep_pydantic=True,
            enable_cache=False,
        )
    ] == [
        dict(field1="value1", field2=42),
        dict(field1="value2", field2=43),
    ]


@pytest.mark.parametrize(
    "func,schema",
    [(batch_generate, SampleModel), (batch_generate_dicts, SampleModel)],
)
def test_cache_behavior(func, schema):
    texts = ["field1: value1, field2: 42", "field1: value2, field2: 43"]

    cache = get_cache("tecton-gen-ai", "structured_outputs")
    mock_cache = Mock(spec=CacheProtocol, wraps=cache)

    with patch(
        "melvin.tecton_gen_ai.utils._cache.get_cache",
        return_value=mock_cache,
    ):
        # Call the function to test
        func("openai/gpt-4o-mini", texts, schema)

        # Ensure cache was checked and set
        assert mock_cache.aget.call_count == 2  # Cache checked twice
        assert mock_cache.aset.call_count == 2  # Cache set twice

        # Call the function again
        func("openai/gpt-4o-mini", texts, schema)

        # Ensure cache was checked again but not set (meaning we returned from cache)
        assert mock_cache.aget.call_count == 4  # Cache checked twice more
        assert mock_cache.aset.call_count == 2  # Cache set should not increase



import asyncio
import os
from typing import Protocol

import diskcache


class CacheProtocol(Protocol):
    def set(self, key, val): ...
    def get(self, key): ...
    async def aset(self, key, val): ...
    async def aget(self, key): ...


class _DiskCache(diskcache.Cache, CacheProtocol):
    async def aset(self, key, val):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.set, key, val)

    async def aget(self, key):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.get, key)


def _cache_dir(*extra_paths: str):
    """Get the cache directory tied to the user

    Example: `~/.cache/$appname/$version`
    """
    path = os.environ.get("XDG_CACHE_HOME", "")
    if not path.strip():
        path = os.path.expanduser("~/.cache")
    return os.path.join(path, *extra_paths)


def get_cache(*extra_paths: str) -> CacheProtocol:
    cache_dir = _cache_dir("tecton-gen-ai", "structured_outputs", *extra_paths)
    # NOTE: default size limit of 1GB (https://grantjenks.com/docs/diskcache/api.html#constants)
    return _DiskCache(cache_dir)



from datetime import timedelta
from typing import Any, List, Optional, Union

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype
from tecton import (
    BatchFeatureView,
    BatchSource,
    Entity,
    PushConfig,
    RealtimeFeatureView,
    RequestSource,
    StreamFeatureView,
    StreamSource,
    batch_feature_view,
    pandas_batch_config,
    realtime_feature_view,
    stream_feature_view,
)
from tecton.types import Field

from ..tecton_utils._internal import (
    get_local_source_attrs,
    set_local_source_attrs,
    set_serialization,
)
from ..utils.config_utils import Configs
from ..tecton_utils._tecton_utils import get_df_schema

_MAX_ROWS = 100
_DEFAULT_TIMESTAMP_FIELD = "_tecton_auto_ts"
_DEFAULT_SOURCE_TIME = "2024-01-01"


def make_local_source(
    name: str,
    raw: Any,
    auto_timestamp: bool = True,
    timestamp_field: Optional[str] = None,
    max_rows: Optional[int] = None,
    is_stream: bool = False,
    **source_kwargs: Any,
) -> Union[BatchSource, StreamSource]:
    """
    Make a local Tecton DataSource from raw data.

    Args:

        name: The name of the source.
        raw: The raw data, can be a pandas DataFrame, a dictionary, or a list of dictionaries.
        auto_timestamp: Whether to automatically add a timestamp field to the data when timestamp_field is None,
            defaults to True.
        timestamp_field: The timestamp field, defaults to None (no timestamp field).
        max_rows: The maximum number of rows allowed in the data, defaults to 100.
        is_stream: Whether the returned source is a stream source, defaults to False.
        **source_kwargs: Other keyword arguments for defining the source.

    Returns:

        A Tecton DataSource.

    Notes:

        - The purpose of this function is to create a local Tecton DataSource for testing and local iteration.
        - It's also deployable to a Tecton cluster. However, it is still for testing purposes only.

    Examples:

        ```python
        from tecton_gen_ai.testing import make_local_source
        import pandas as pd

        # simplest form
        my_src1 = make_local_source(
            name="my_src1",
            raw={"user_id": [1, 2]}),
            description="User information",
        )

        # with a list of dictionaries
        data = [
            {"user_id": 1, "event_time": "2024-01-01T00:00:00", "event_type": "click"},
            {"user_id": 2, "event_time": "2024-01-01T00:00:01", "event_type": "click"},
        ]
        my_src2 = make_local_source(
            name="my_src2",
            raw=data,
            timestamp_field="event_time",
            description="User information",
        )

        # with a pandas DataFrame
        df = pd.DataFrame(data)
        my_src3 = make_local_source(
            name="my_src3",
            raw=df,
            timestamp_field="event_time",
            description="User information",
        )
        ```
    """
    df = _to_df(raw)
    _max_row = max_rows or _MAX_ROWS
    if len(df) > _max_row:
        raise ValueError(f"Dataframe has more than {_max_row} rows")

    if timestamp_field is None and auto_timestamp:
        timestamp_field: Optional[str] = _DEFAULT_TIMESTAMP_FIELD
        df = df.assign(**{timestamp_field: _DEFAULT_SOURCE_TIME})
    if timestamp_field is not None:
        df = df.assign(**{timestamp_field: pd.to_datetime(df[timestamp_field])})

    normalized_cols = {}
    ts_fields = []
    for col in df.columns:
        cv = df[col]
        if is_datetime64_any_dtype(cv):
            cv = cv.astype(str)
            ts_fields.append(col)
        normalized_cols[col] = cv

    data = pd.DataFrame(normalized_cols).to_dict("records")

    with set_serialization():

        @pandas_batch_config(supports_time_filtering=True)
        def api_df(filter_context):
            import pandas as pd

            df = pd.DataFrame(data)
            for f in ts_fields:
                df[f] = pd.to_datetime(df[f])
            return df

    if not is_stream:
        src = BatchSource(name=name, batch_config=api_df, **source_kwargs)
    else:
        src = StreamSource(
            name=name,
            stream_config=PushConfig(),
            batch_config=api_df,
            schema=get_df_schema(df),
            **source_kwargs,
        )

    mock_params = {
        "timestamp_field": timestamp_field,
    }
    if timestamp_field is not None:
        mock_params["start_time"] = df[timestamp_field].min().to_pydatetime()
        mock_params["end_time"] = df[timestamp_field].max().to_pydatetime() + timedelta(
            days=1
        )
    set_local_source_attrs(src, mock_params)
    return src


def make_local_batch_feature_view(
    name: str,
    data: Any,
    entity_keys: List[str],
    timestamp_field: Optional[str] = None,
    max_rows: Optional[int] = None,
    **fv_kwargs: Any,
) -> BatchFeatureView:
    """
    Make a local Tecton BatchFeatureView from raw data. Under the hood, it will call `make_local_source`
    to create a local source.

    Args:

        name: The name of the feature view.
        data: The raw data, can be a pandas DataFrame, a dictionary, or a list of dictionaries.
        entity_keys: The entity keys for dedup and lookup.
        timestamp_field: The timestamp field, defaults to None (will add timestamp field automatically).
        max_rows: The maximum number of rows allowed in the data, defaults to None.
        **fv_kwargs: Other keyword arguments for defining the feature view.

    Returns:

        A Tecton BatchFeatureView.

    Notes:

        - The purpose of this function is to create a local Tecton BatchFeatureView for testing and local iteration.
        - It's also deployable to a Tecton cluster. However, it is still for testing purposes only.
        - Adding description to the feature view is strongly recommended (for LLMs to understand the feature view).

    Examples:

        ```python
        from tecton_gen_ai.testing import make_local_batch_feature_view

        # simplest form
        my_fv1 = make_local_batch_feature_view(
            "my_fv1",
            {"user_id": 1, "name": "Jim"},
            ["user_id"],
            "description": "User information",
        )

        # with a list of dictionaries
        data = [
            {"user_id": 1, "name": "Jim"},
            {"user_id": 2, "name": "John"},
        ]
        my_fv2 = make_local_batch_feature_view(
            "my_fv2", data, ["user_id"],
            description="User information",
        )

        # with a pandas DataFrame
        import pandas as pd

        df = pd.DataFrame(data)
        my_fv3 = make_local_batch_feature_view(
            "my_fv3", df, ["user_id"],
            description="User information",
        )
        ```
    """
    df = _to_df(data)
    source = make_local_source(
        name + "_source",
        df,
        max_rows=max_rows,
        timestamp_field=timestamp_field,
        auto_timestamp=True,
    )
    timestamp_field = get_local_source_attrs(source)["timestamp_field"]
    start = get_local_source_attrs(source)["start_time"]
    schema = get_df_schema(df, as_attributes=True)
    join_keys = [Field(x.name, x.dtype) for x in schema if x.name in entity_keys]
    if len(join_keys) != len(entity_keys):
        raise ValueError(f"Entity keys {entity_keys} not all found in schema {schema}")
    entity = Entity(name=name + "_entity", join_keys=join_keys)
    features = fv_kwargs.get(
        "features",
        [x for x in schema if x.name not in entity_keys and x.name != timestamp_field],
    )

    base_args = dict(
        name=name,
        sources=[source],
        entities=[entity],
        mode="pandas",
        online=True,
        offline=True,
        features=features,
        feature_start_time=start,
        incremental_backfills=False,
        batch_schedule=timedelta(days=1),
        max_backfill_interval=timedelta(days=10000),
        timestamp_field=timestamp_field,
        **Configs.get_default().get_bfv_config(),
    )
    base_args.update(fv_kwargs)

    with set_serialization():

        @batch_feature_view(**base_args)
        def dummy(_df):
            return _df

    return dummy


def make_local_realtime_feature_view(
    name: str,
    data: Any,
    entity_keys: List[str],
    **fv_kwargs: Any,
) -> RealtimeFeatureView:
    """
    Make a local Tecton RealtimeFeatureView from raw data.

    Args:

        name: The name of the feature view.
        data: The raw data, can be a pandas DataFrame, a dictionary, or a list of dictionaries.
        entity_keys: The entity keys for lookup.
        **fv_kwargs: Other keyword arguments for defining the feature view.

    Returns:

        A Tecton RealtimeFeatureView.

    Notes:

        - The purpose of this function is to create a local Tecton RealtimeFeatureView for testing and local iteration.
        - It's also deployable to a Tecton cluster. However, it is still for testing purposes only.

    Examples:

        ```python
        from tecton_gen_ai.testing import make_local_realtime_feature_view

        fv = make_local_realtime_feature_view("fv", {"user_id": "user1", "age": 30}, ["user_id"])
        events = pd.DataFrame([{"user_id": "user1"}])
        odf = fv.get_features_for_events(events).to_pandas()
        ```
    """
    df = _to_df(data)
    request_source = RequestSource(schema=get_df_schema(df[entity_keys]))
    features = fv_kwargs.get("features", get_df_schema(df, as_attributes=True))

    base_args = dict(
        name=name,
        sources=[request_source],
        mode="pandas",
        features=features,
        **Configs.get_default().get_rtfv_config(),
    )
    base_args.update(fv_kwargs)

    with set_serialization():

        @realtime_feature_view(**base_args)
        def dummy(request):
            import pandas as pd

            return pd.merge(request, df, on=entity_keys, how="inner").head(1)

    return dummy


def make_local_stream_feature_view(
    name: str,
    data: Any,
    entity_keys: List[str],
    timestamp_field: Optional[str] = None,
    max_rows: Optional[int] = None,
    **fv_kwargs: Any,
) -> StreamFeatureView:
    """
    Make a local Tecton StreamFeatureView from raw data.

    Args:

        name: The name of the feature view.
        data: The raw data, can be a pandas DataFrame, a dictionary, or a list of dictionaries.
        entity_keys: The entity keys for lookup.
        timestamp_field: The timestamp field, defaults to None (will add timestamp field automatically).
        max_rows: The maximum number of rows allowed in the data, defaults to None.
        **fv_kwargs: Other keyword arguments for defining the feature view.

    Returns:

        A Tecton StreamFeatureView.

    Notes:

        - The purpose of this function is to create a local Tecton StreamFeatureView for testing and local iteration.
        - It's also deployable to a Tecton cluster. However, it is still for testing purposes only.

    Examples:

        ```python
        from tecton_gen_ai.testing import make_local_stream_feature_view

        # simplest form
        fv = make_local_stream_feature_view("fv", {"user_id": "user1", "age": 30}, ["user_id"])

        # with a list of dictionaries
        data = [
            {"user_id": "user1", "age": 30},
            {"user_id": "user2", "age": 40},
        ]
        fv = make_local_stream_feature_view("fv", data, ["user_id"])

        # with a pandas DataFrame
        import pandas as pd

        df = pd.DataFrame(data)
        fv = make_local_stream_feature_view("fv", df, ["user_id"])
        ```
    """
    df = _to_df(data)
    source = make_local_source(
        name + "_source",
        df,
        max_rows=max_rows,
        is_stream=True,
        auto_timestamp=True,
        timestamp_field=timestamp_field,
    )
    timestamp_field = get_local_source_attrs(source)["timestamp_field"]
    start = get_local_source_attrs(source)["start_time"]
    schema = get_df_schema(df, as_attributes=True)
    join_keys = [Field(x.name, x.dtype) for x in schema if x.name in entity_keys]
    if len(join_keys) != len(entity_keys):
        raise ValueError(f"Entity keys {entity_keys} not all found in schema {schema}")
    entity = Entity(name=name + "_entity", join_keys=join_keys)
    features = fv_kwargs.get(
        "features",
        [x for x in schema if x.name not in entity_keys and x.name != timestamp_field],
    )

    base_args = dict(
        name=name,
        source=source,
        entities=[entity],
        mode="pandas",
        online=True,
        offline=True,
        features=features,
        feature_start_time=start,
        batch_schedule=timedelta(days=1),
        max_backfill_interval=timedelta(days=10000),
        timestamp_field=timestamp_field,
        **Configs.get_default().get_sfv_config(),
    )
    base_args.update(fv_kwargs)

    with set_serialization():

        @stream_feature_view(**base_args)
        def dummy(_df):
            return _df

    return dummy


def _to_df(data: Any) -> pd.DataFrame:
    if isinstance(data, pd.DataFrame):
        return data
    if isinstance(data, dict):
        return pd.DataFrame([data])
    if isinstance(data, list):
        return pd.DataFrame(data)
    raise ValueError(f"Unsupported data type {type(data)}")



from pytest import fixture

from melvin.tecton_gen_ai.api import Agent, prompt
from melvin.tecton_gen_ai.testing import make_local_batch_feature_view
from pydantic import BaseModel, Field


@fixture
def mock_data():
    return [
        {
            "user_id": "user1",
            "name": "Jim",
            "age": 30,
            "food_preference": "American",
        },
        {
            "user_id": "user2",
            "name": "John",
            "age": 40,
            "food_preference": "Italian",
        },
        {
            "user_id": "user3",
            "name": "Jane",
            "age": 50,
            "food_preference": "Chinese",
        },
    ]


@fixture
def mock_agent_service(set_default_llm, mock_data, mock_knowledge):
    class Output(BaseModel):
        number: int = Field(description="The number to the question")

    user_info = make_local_batch_feature_view(
        "user_info", mock_data, entity_keys=["user_id"], description="User info"
    )

    @prompt(sources=[user_info])
    def sys_prompt(location: str, user_info):
        name = user_info["name"]
        return f"You are serving {name} in {location}"

    def get_tecton_employee_count() -> int:
        """
        Returns the number of employees in Tecton
        """
        return 110

    return Agent(
        name="test",
        description="Test agent",
        prompt=sys_prompt,
        tools=[user_info, get_tecton_employee_count],
        knowledge=[mock_knowledge],
        output_schema=Output,
    )


def test_agent(mock_agent_service):
    client = mock_agent_service
    assert "You are serving Jane in Chicago" == client.invoke_prompt(
        dict(user_id="user3", location="Chicago")
    )
    sys_prompt = client.invoke_system_prompt(dict(user_id="user3", location="Chicago"))
    assert "You are serving Jane in Chicago" in sys_prompt
    assert "Test agent" in sys_prompt
    assert "context" in sys_prompt.lower()
    assert client.invoke_tool("user_info", dict(user_id="user1")) == {
        "name": "Jim",
        "age": 30,
        "food_preference": "American",
    }
    assert client.invoke_tool("get_tecton_employee_count") == 110

    assert client.search("knowledge", query="food", filter={"zip": 98010}) == []
    assert (
        len(client.search("knowledge", query="food", top_k=5, filter={"zip": 98005}))
        == 3
    )
    assert len(client.search("knowledge", query="food", top_k=5)) == 5

    assert client.invoke(
        "how many employees in tecton",
        context=dict(user_id="user3", location="Chicago"),
    ) == {"number": 110}



import inspect
import json
from datetime import datetime
from functools import lru_cache
from textwrap import dedent
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Type, Union

from pydantic import BaseModel, Field, create_model


def serialize_pydantic(obj: Type[BaseModel]) -> str:
    return json.dumps(to_openai_function(obj))


@lru_cache
def deserialize_pydantic(model_json: str) -> Type[BaseModel]:
    return openai_function_to_pydantic(json.loads(model_json))


def to_openai_function(
    obj: Union[Callable, Type[BaseModel]],
    name: Optional[str] = None,
    exclude_args: Optional[Iterator[str]] = None,
    description: Optional[str] = None,
) -> Dict[str, Any]:
    exclude = set(exclude_args or [])

    if inspect.isclass(obj) and issubclass(obj, BaseModel):
        from openai import pydantic_function_tool

        res = dict(pydantic_function_tool(obj))["function"]
    else:
        from langchain_core.utils.function_calling import convert_to_openai_function

        res = convert_to_openai_function(obj)
    res["name"] = name or res["name"]
    if description is not None or obj.__doc__ is not None:
        res["description"] = description or dedent(obj.__doc__)
    properties = res["parameters"]["properties"]
    for key in list(properties.keys()):
        if key in exclude:
            del properties[key]
    return res


def openai_function_to_pydantic(
    model: Dict[str, Any], name: Optional[str] = None
) -> Type[BaseModel]:
    return openai_properties_to_pydantic(
        model["parameters"]["properties"],
        name=name or model["name"],
        description=model.get("description"),
    )


def to_openai_properties(
    func: Union[Callable, Type[BaseModel], List[Tuple[str, Any]]], args: List[str]
) -> Dict[str, Any]:
    if inspect.isclass(func) and issubclass(func, BaseModel):
        new_fields = {
            k: (v.annotation, v) for k, v in func.model_fields.items() if k in args
        }
        model = create_model("temp", **new_fields)
    else:
        from ..tecton_utils.convert import parse_python_annotation

        if isinstance(func, list):
            annotations = func
        else:
            annotations = list(func.__annotations__.items())
        fields: Dict[str, Any] = {}
        for k, v in annotations:
            if k not in args:
                continue
            v, field = parse_python_annotation(v)
            fields[k] = (v, field)
        model = create_model("temp", **fields)
    res = to_openai_function(model)
    return res["parameters"]["properties"]


def openai_properties_to_pydantic(
    properties: Dict[str, Any],
    name: Optional[str] = None,
    description: Optional[str] = None,
) -> Type[BaseModel]:
    fields = {}
    for name, value in properties.items():
        dtype = openai_type_to_python_type(value)
        fparams = {}
        if "description" in value:
            fparams["description"] = value["description"]
        if "default" in value:
            fparams["default"] = value["default"]
        fields[name] = (dtype, Field(**fparams))
    return create_model(name or "temp", __doc__=description, **fields)


def openai_type_to_python_type(tp: Dict[str, Any]) -> Any:
    if "anyOf" in tp:
        tps = [openai_type_to_python_type(t) for t in tp["anyOf"]]
        return Union[tuple(tps)]
    value = tp.get("type")
    if value == "integer":
        return int
    if value == "number":
        return float
    if value == "boolean":
        return bool
    if value == "string":
        format = tp.get("format", "")
        if format == "":
            return str
        if format == "date-time":
            return datetime
        raise NotImplementedError(f"Unsupported string format {format}")
    if value == "null":
        return None
    if value == "array":
        if "items" in tp:
            items = tp["items"]
            return List[openai_type_to_python_type(items)]
        if "prefixItems" in tp:
            pitems = [openai_type_to_python_type(t) for t in tp["prefixItems"]]
            return Tuple[tuple(pitems)]
    if value == "object":
        additional = tp.get("additionalProperties")
        if additional is not None:
            return Dict[str, openai_type_to_python_type(additional)]
    raise NotImplementedError(f"Unsupported openai data type {tp}")


def _annotation(_type: Any) -> str:
    if _type is None:
        return "None"
    if _type in (str, int, float, bool):
        return _type.__name__
    if _type is Any:
        return "Any"
    if _type.__origin__ is list:
        return f"List[{_annotation(_type.__args__[0])}]"
    if _type.__origin__ is dict:
        return (
            f"Dict[{_annotation(_type.__args__[0])}, {_annotation(_type.__args__[1])}]"
        )
    raise NotImplementedError(f"Unsupported annotation ype {_type}")



from pydantic import BaseModel, Field
from pytest import fixture

from melvin.tecton_gen_ai.api import llm_extraction
from melvin.tecton_gen_ai.testing import make_local_source
from melvin.tecton_gen_ai.tecton_utils._internal import get_local_source_attrs
from melvin.tecton_gen_ai.tecton_utils._tecton_utils import make_entity


@fixture
def mock_source():
    return make_local_source(
        "name_and_age",
        [
            {"s1": "my name is Jim", "s2": "my age is 20"},
            {"s1": "my name is John", "s2": "my age is 30"},
        ],
        description="User info",
    )


@fixture
def extraction_config():
    class ExtractName(BaseModel):
        name: str = Field(description="mentioned name")

    class ExtractAge(BaseModel):
        age: int = Field(description="mentioned age")

    return [
        {
            "column": "s1",
            "schema": ExtractName,
            "enable_local_cache": False,
        },
        {
            "column": "s2",
            "schema": ExtractAge,
            "enable_local_cache": False,
        },
    ]


def test_llm_extraction(set_default_llm, extraction_config, mock_source):
    entity = make_entity(name=str)
    e = llm_extraction(
        source=mock_source, extraction_config=extraction_config, entities=[entity]
    )
    attrs = get_local_source_attrs(mock_source)
    df = e.run_transformation(
        start_time=attrs["start_time"], end_time=attrs["end_time"]
    ).to_pandas()
    assert df[["name", "age"]].to_dict(orient="records") == [
        {"name": "Jim", "age": 20},
        {"name": "John", "age": 30},
    ]



from typing import Optional

import pandas as pd
import snowflake.connector
from utils import _err, _log_with_details, store_object


class SnowflakeCostQueryTool:
    def __init__(self):
        self.connection = None

    def connect(self, user: str):
        self.connection = self.connection or snowflake.connector.connect(
            account="hza50686",
            user=user,
            warehouse="ANALYST_WAREHOUSE",
            database="DATAMART_PROD",
            schema="CONSUMPTION",
            authenticator="externalbrowser",  # Uses Okta login
        )

    def build_query(
        self,
        account_name: Optional[str],
        timestamp_start: Optional[str],
        timestamp_end: Optional[str],
    ) -> str:
        """
        Build the SQL query based on the input filters and groupings.
        """
        base_query = """
        SELECT
            ACCOUNT_NAME,
            UNIT,
            SUM(QUANTITY) AS QUANTITY,
            COALESCE(metadata:instance_type::string, 'Unknown') AS Instance_Type,
            COALESCE(metadata:job_type::string, 'Unknown') AS Job_Type,
            COALESCE(metadata:tecton_object_name::string, 'Unknown') AS FV_Name,
            COALESCE(metadata:workspace::string, 'Unknown') AS Workspace,
            TO_CHAR(TIMESTAMP, 'YYYY-MM-DD') AS TIMESTAMP,
            BASE_TECTON_CREDITS_PER_UNIT,
            SUM(QUANTITY) * BASE_TECTON_CREDITS_PER_UNIT AS TOTAL_TECTON_CREDITS
        FROM
            COMBINED_CONSUMPTION_METRICS_WITH_CREDITS_V2
        """

        # Add WHERE clause for filters
        where_clauses = []
        if account_name:
            where_clauses.append(f"ACCOUNT_NAME = '{account_name}'")
        if timestamp_start and timestamp_end:
            where_clauses.append(
                f"TIMESTAMP >= '{timestamp_start}' AND TIMESTAMP < '{timestamp_end}'"
            )

        if where_clauses:
            base_query += "WHERE " + " AND ".join(where_clauses) + " "

        # Add GROUP BY clause
        group_by_fields = [
            "ACCOUNT_NAME",
            "Workspace",
            "UNIT",
            "FV_Name",
            "Job_Type",
            "Instance_Type",
            "TO_CHAR(TIMESTAMP, 'YYYY-MM-DD')",
            "BASE_TECTON_CREDITS_PER_UNIT",
        ]

        base_query += "GROUP BY " + ", ".join(group_by_fields) + " "

        # Add ORDER BY clause
        base_query += "ORDER BY TIMESTAMP, UNIT, QUANTITY DESC"

        return base_query

    def execute_query(self, query: str) -> pd.DataFrame:
        """
        Execute the SQL query and fetch results.
        """
        cursor = self.connection.cursor()
        try:
            _log_with_details(":question: Fetching cost information", f"```sql {query}")

            cursor.execute(query)
            df = cursor.fetch_pandas_all()
            df["TIMESTAMP"] = pd.to_datetime(df["TIMESTAMP"]).astype("datetime64[us]")
            return df
        finally:
            cursor.close()

    def query_cost_data(
        self,
        user: str,
        account_name: str,
        timestamp_start: Optional[str] = None,
        timestamp_end: Optional[str] = None,
    ) -> str:
        """
        Query Tecton cost data. This is a dataframe generator.

        Parameters:
        - user: The full email address of the currently logged in user
        - account_name: The name of the currently logged into Tecton account to filter by (e.g., "tecton-staging").
        - timestamp_start: Start of the timestamp range (e.g., "2024-05-01").
        - timestamp_end: End of the timestamp range

        Returns:
            - str: A session object key of the cost data or an error message

        Note:

        - The output columns are TIMESTAMP, INSTANCE_TYPE, JOB_TYPE, FV_NAME, WORKSPACE, TOTAL_TECTON_CREDITS
        """
        if not timestamp_start or not timestamp_end:
            return _err("Both timestamp_start and timestamp_end must be provided.")

        try:
            self.connect(user)

            query = self.build_query(
                account_name=account_name,
                timestamp_start=timestamp_start,
                timestamp_end=timestamp_end,
            )
            df = self.execute_query(query)[
                [
                    "TIMESTAMP",
                    "INSTANCE_TYPE",
                    "JOB_TYPE",
                    "FV_NAME",
                    "WORKSPACE",
                    "TOTAL_TECTON_CREDITS",
                ]
            ]
            description = """
Schema of the cost table: (format: name (type): description)

TIMESTAMP (timestamp): timestamp of the job
INSTANCE_TYPE (string): cloud instance type
JOB_TYPE (string): job type
WORKSPACE (string): workspace name
FV_NAME (string): feature view name
TOTAL_TECTON_CREDITS (float): total credits consumed by the job
"""
            return store_object({"df": df, "description": description}, "df")
        except Exception as e:
            return _err(e)
